# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################


import os
from urllib.parse import urlencode

import requests
from requests import HTTPError

import sickrage
from sickrage.core.websession import WebSession
from sickrage.notifiers import Notifiers


class pyTivoNotifier(Notifiers):
    def __init__(self):
        super(pyTivoNotifier, self).__init__()
        self.name = 'pytivo'

    def notify_snatch(self, ep_name):
        pass

    def notify_download(self, ep_name):
        pass

    def notify_subtitle_download(self, ep_name, lang):
        pass

    def notify_version_update(self, new_version):
        pass

    def update_library(self, ep_obj):

        # Values from config

        if not sickrage.app.config.use_pytivo:
            return False

        host = sickrage.app.config.pytivo_host
        share_name = sickrage.app.config.pytivo_share_name
        tsn = sickrage.app.config.pytivo_tivo_name

        # There are two more values required, the container and file.
        #
        # container: The share name, show name and season
        #
        # file: The file name
        #
        # Some slicing and dicing of variables is required to get at these values.
        #
        # There might be better ways to arrive at the values, but this is the best I have been able to
        # come up with.
        #

        # Calculated values

        show_path = ep_obj.show.location
        show_name = ep_obj.show.name
        root_show_and_season = os.path.dirname(ep_obj.location)
        abs_path = ep_obj.location

        # Some show names have colons in them which are illegal in a path location, so strip them out.
        # (Are there other characters?)
        show_name = show_name.replace(":", "")

        root = show_path.replace(show_name, "")
        show_and_season = root_show_and_season.replace(root, "")

        container = share_name + "/" + show_and_season
        file = "/" + abs_path.replace(root, "")

        # Finally create the url and make request
        request_url = "http://{}/TiVoConnect?{}".format(host, urlencode({'Command': 'Push', 'Container': container, 'File': file, 'tsn': tsn}))

        sickrage.app.log.debug("pyTivo notification: Requesting " + request_url)

        try:
            WebSession().get(request_url)
        except requests.exceptions.HTTPError as e:
            sickrage.app.log.error("pyTivo notification: Error, the server couldn't fulfill the request - " + e.response.text)
            return False
        except Exception as e:
            sickrage.app.log.error("PYTIVO: Unknown exception: {}".format(e))
            return False

        sickrage.app.log.info("pyTivo notification: Successfully requested transfer of file")
        return True



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.

from sqlalchemy import Column, Integer, Text, String, Boolean
from sqlalchemy.ext.declarative import as_declarative

from sickrage.core.databases import SRDatabase, SRDatabaseBase


@as_declarative()
class CacheDBBase(SRDatabaseBase):
    pass


class CacheDB(SRDatabase):
    def __init__(self, db_type, db_prefix, db_host, db_port, db_username, db_password):
        super(CacheDB, self).__init__('cache', 7, db_type, db_prefix, db_host, db_port, db_username, db_password)
        CacheDBBase.metadata.create_all(self.engine)
        for model in CacheDBBase._decl_class_registry.values():
            if hasattr(model, '__tablename__'):
                self.tables[model.__tablename__] = model

    def cleanup(self):
        def remove_duplicates_from_last_search_table():
            found = []

            session = self.session()

            for x in session.query(CacheDB.LastSearch).all():
                if x.provider in found:
                    x.delete()
                    session.commit()
                else:
                    found.append(x.provider)

        def remove_duplicates_from_scene_name_table():
            found = []

            session = self.session()

            for x in session.query(CacheDB.SceneName).all():
                if (x.indexer_id, x.name) in found:
                    x.delete()
                    session.commit()
                else:
                    found.append((x.indexer_id, x.name))

        remove_duplicates_from_last_search_table()
        # remove_duplicates_from_scene_name_table()

    class LastUpdate(CacheDBBase):
        __tablename__ = 'last_update'

        provider = Column(String(32), primary_key=True)
        time = Column(Integer)

    class LastSearch(CacheDBBase):
        __tablename__ = 'last_search'

        provider = Column(String(32), primary_key=True)
        time = Column(Integer)

    class SceneName(CacheDBBase):
        __tablename__ = 'scene_names'

        id = Column(Integer, primary_key=True)
        indexer_id = Column(Integer)
        name = Column(Text)

    class NetworkTimezone(CacheDBBase):
        __tablename__ = 'network_timezones'

        network_name = Column(String(256), primary_key=True)
        timezone = Column(Text)

    class Provider(CacheDBBase):
        __tablename__ = 'providers'

        id = Column(Integer, primary_key=True)
        provider = Column(Text)
        name = Column(Text)
        season = Column(Integer)
        episodes = Column(Text)
        series_id = Column(Integer)
        url = Column(String(256), index=True, unique=True)
        time = Column(Integer)
        quality = Column(Integer)
        release_group = Column(Text)
        version = Column(Integer, default=-1)
        seeders = Column(Integer)
        leechers = Column(Integer)
        size = Column(Integer)

    class QuickSearchShow(CacheDBBase):
        __tablename__ = 'quicksearch_shows'

        category = Column(Text)
        showid = Column(Integer, index=True, primary_key=True)
        seasons = Column(Integer)
        name = Column(Text)
        img = Column(Text)

    class QuickSearchEpisode(CacheDBBase):
        __tablename__ = 'quicksearch_episodes'

        category = Column(Text)
        showid = Column(Integer, index=True, primary_key=True)
        episodeid = Column(Integer)
        season = Column(Integer, index=True, primary_key=True)
        episode = Column(Integer, index=True, primary_key=True)
        name = Column(Text)
        showname = Column(Text)
        img = Column(Text)

    class OAuth2Token(CacheDBBase):
        __tablename__ = 'oauth2_token'

        id = Column(Integer, primary_key=True)
        access_token = Column(String(255), unique=True, nullable=False)
        refresh_token = Column(String(255), index=True)
        expires_in = Column(Integer, nullable=False, default=0)
        expires_at = Column(Integer, nullable=False, default=0)
        scope = Column(Text, default="")
        session_state = Column(Text, default="")
        token_type = Column(Text, default="bearer")

    class Announcements(CacheDBBase):
        __tablename__ = 'announcements'

        id = Column(Integer, primary_key=True)
        hash = Column(String(255), unique=True, nullable=False)
        seen = Column(Boolean, default=False)



# coding=utf-8
# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
# Git: https://git.sickrage.ca/SiCKRAGE/sickrage
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.


import re

import sickrage
from sickrage.core.caches.tv_cache import TVCache
from sickrage.core.helpers import bs4_parser, try_int, convert_size
from sickrage.providers import TorrentProvider


class YggtorrentProvider(TorrentProvider):
    def __init__(self):
        """Initialize the class."""
        super(YggtorrentProvider, self).__init__('Yggtorrent', 'https://www2.yggtorrent.se', True)

        # URLs
        self._urls.update({
            'auth': '{base_url}/user/ajax_usermenu'.format(**self._urls),
            'login': '{base_url}/user/login'.format(**self._urls),
            'search': '{base_url}/engine/search'.format(**self._urls),
            'download': '{base_url}/engine/download_torrent?id=%s'.format(**self._urls)
        })

        # Credentials
        self.username = None
        self.password = None

        # Proper Strings
        self.proper_strings = ['PROPER', 'REPACK', 'REAL', 'RERIP']

        # Torrent Stats
        self.minseed = None
        self.minleech = None

        # Cache
        self.cache = TVCache(self, min_time=20)

    def search(self, search_strings, age=0, show_id=None, season=None, episode=None, **kwargs):
        """
        Search a provider and parse the results.

        :param search_strings: A dict with mode (key) and the search value (value)
        :param age: Not used
        :param ep_obj: Not used
        :returns: A list of search results (structure)
        """
        results = []

        if not self.login():
            return results

        # Search Params
        search_params = {
            'category': 2145,
            'do': 'search'
        }

        for mode in search_strings:
            sickrage.app.log.debug('Search mode: {}'.format(mode))
            for search_string in search_strings[mode]:
                if mode != 'RSS':
                    sickrage.app.log.debug('Search string: {}'.format(search_string))

                search_params['name'] = re.sub(r'[()]', '', search_string)

                resp = self.session.get(self.urls['search'], params=search_params)
                if not resp or not resp.text:
                    sickrage.app.log.debug('No data returned from provider')
                    continue

                results += self.parse(resp.text, mode)

        return results

    def parse(self, data, mode, **kwargs):
        """
        Parse search results for items.

        :param data: The raw response from a search
        :param mode: The current mode used to search, e.g. RSS

        :return: A list of items found
        """
        results = []

        with bs4_parser(data) as html:
            torrent_table = html.find(class_='table-responsive results')
            torrent_rows = torrent_table('tr') if torrent_table else []

            # Continue only if at least one Release is found
            if len(torrent_rows) < 2:
                sickrage.app.log.debug('Data returned from provider does not contain any torrents')
                return results

            for result in torrent_rows[1:]:
                cells = result('td')
                if len(cells) < 9:
                    continue

                try:
                    info = cells[1].find('a')
                    title = info.get_text(strip=True)
                    download_url = info.get('href')
                    if not (title and download_url):
                        continue

                    torrent_id = re.search(r'/(\d+)-', download_url)
                    download_url = self.urls['download'] % torrent_id.group(1)

                    seeders = try_int(cells[7].get_text(strip=True), 0)
                    leechers = try_int(cells[8].get_text(strip=True), 0)

                    torrent_size = cells[5].get_text()
                    size = convert_size(torrent_size, -1, ['O', 'KO', 'MO', 'GO', 'TO', 'PO'])

                    results += [{
                        'title': title,
                        'link': download_url,
                        'size': size,
                        'seeders': seeders,
                        'leechers': leechers
                    }]

                    if mode != 'RSS':
                        sickrage.app.log.debug("Found result: {}".format(title))
                except Exception:
                    sickrage.app.log.error('Failed parsing provider.')

        return results

    def login(self):
        """Login method used for logging in before doing search and torrent downloads."""
        login_params = {
            'id': self.username,
            'pass': self.password
        }

        if not self._is_authenticated():
            try:
                self.session.post(self.urls['login'], data=login_params, allow_post_redirects=True)
            except Exception:
                sickrage.app.log.warning('Unable to connect or login to provider')
                return False

            if not self._is_authenticated():
                sickrage.app.log.warning('Invalid username or password. Check your settings')
                return False

        return True

    def _is_authenticated(self):
        try:
            self.session.get(self.urls['auth']).json()
        except Exception:
            return False
        return True



#!/usr/bin/env python
# coding=utf-8
#
# This file is part of aDBa.
#
# aDBa is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# aDBa is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with aDBa.  If not, see <http://www.gnu.org/licenses/>.

import logging
import socket
import sys
import threading
import zlib
from builtins import bytes
from time import time, sleep

from .aniDBerrors import *
from .aniDBresponses import ResponseResolver

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class AniDBLink(threading.Thread):
    def __init__(self, server, port, myport, delay=2, timeout=20):
        super(AniDBLink, self).__init__()
        self.server = server
        self.port = port
        self.target = (server, port)
        self.timeout = timeout

        self.myport = 0
        self.bound = self.connectSocket(myport, self.timeout)

        self.cmd_queue = {None: None}
        self.resp_tagged_queue = {}
        self.resp_untagged_queue = []
        self.tags = []
        self.lastpacket = time()
        self.delay = delay
        self.session = None
        self.banned = False
        self.crypt = None

        self._stop = threading.Event()
        self._quiting = False

        self.QuitProcessed = False

        self.setDaemon(True)
        self.start()

    def connectSocket(self, myport, timeout):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.settimeout(timeout)
        portlist = [myport] + [7654]
        for port in portlist:
            try:
                self.sock.bind(('', port))
            except:
                continue
            else:
                self.myport = port
                return True
        else:
            return False

    def disconnectSocket(self):
        self.sock.shutdown(socket.SHUT_RD)
        # close is not called as the garbage collection from python will handle this for us.  Calling close can also cause issues with the threaded code.
        # self.sock.close()

    def stop(self):
        logger.info("Releasing socket and stopping link thread")
        self._quiting = True
        self.disconnectSocket()
        self._stop.set()

    def stopped(self):
        return self._stop.isSet()

    def print_log_dummy(self, data):
        pass

    def run(self):
        while not self._quiting:
            try:
                data = self.sock.recv(8192)
            except socket.timeout:
                self._handle_timeouts()
                continue
            except OSError as error:
                logger.exception('Exception: %s', error)
                break
            logger.debug("NetIO < %r", data)
            try:
                for i in range(2):
                    try:
                        tmp = data
                        resp = None
                        if tmp[:2] == b'\x00\x00':
                            tmp = zlib.decompressobj().decompress(tmp[2:])
                            logger.debug("UnZip | %r", tmp)
                        resp = ResponseResolver(tmp)
                    except Exception as e:
                        logger.exception('Exception: %s', e)
                        sys.excepthook(*sys.exc_info())
                        self.crypt = None
                        self.session = None
                    else:
                        break
                if not resp:
                    raise AniDBPacketCorruptedError("Either decrypting, decompressing or parsing the packet failed")
                cmd = self._cmd_dequeue(resp)
                resp = resp.resolve(cmd)
                resp.parse()
                if resp.rescode in (b'200', b'201'):
                    self.session = resp.attrs[b'sesskey']
                if resp.rescode in (b'209',):
                    logger.error("sorry encryption is not supported")
                    raise AniDBError()
                    # self.crypt=aes(md5(resp.req.apipassword+resp.attrs['salt']).digest())
                if resp.rescode in (b'203', b'403', b'500', b'501', b'503', b'506'):
                    self.session = None
                    self.crypt = None
                if resp.rescode in (b'504', b'555'):
                    self.banned = True
                    logger.critical(("AniDB API informs that user or client is banned:", resp.resstr))
                resp.handle()
                if not cmd or not cmd.mode:
                    self._resp_queue(resp)
                else:
                    self.tags.remove(resp.restag)
            except:
                sys.excepthook(*sys.exc_info())
                logger.error("Avoiding flood by paranoidly panicing: Aborting link thread, killing connection, releasing waiters and quiting")
                self.sock.close()
                try:
                    cmd.waiter.release()
                except:
                    pass
                for tag, cmd in self.cmd_queue.items():
                    try:
                        cmd.waiter.release()
                    except:
                        pass
                sys.exit()
        if self._quiting:
            self.QuitProcessed = True

    def _handle_timeouts(self):
        willpop = []
        for tag, cmd in self.cmd_queue.items():
            if not tag:
                continue
            if time() - cmd.started > self.timeout:
                self.tags.remove(cmd.tag)
                willpop.append(cmd.tag)
                cmd.waiter.release()

        for tag in willpop:
            self.cmd_queue.pop(tag)

    def _resp_queue(self, response):
        if response.restag:
            self.resp_tagged_queue[response.restag] = response
        else:
            self.resp_untagged_queue.append(response)

    def getresponse(self, command):
        if command:
            resp = self.resp_tagged_queue.pop(command.tag)
        else:
            resp = self.resp_untagged_queue.pop()
        self.tags.remove(resp.restag)
        return resp

    def _cmd_queue(self, command):
        self.cmd_queue[command.tag] = command
        self.tags.append(command.tag)

    def _cmd_dequeue(self, resp):
        if not resp.restag:
            return None
        else:
            return self.cmd_queue.pop(resp.restag)

    def _delay(self):
        return self.delay < 2.1 and 2.1 or self.delay

    def _do_delay(self):
        age = time() - self.lastpacket
        delay = self._delay()
        if age <= delay:
            sleep(delay - age)

    def _send(self, command):
        if self.banned:
            logger.debug("NetIO | BANNED")
            raise AniDBError("Not sending, banned")
        self._do_delay()
        self.lastpacket = time()
        command.started = time()
        data = command.raw_data()

        self.sock.sendto(bytes(data, "ASCII"), self.target)
        if command.command == b'AUTH':
            logger.debug("NetIO > sensitive data is not logged!")

    def new_tag(self):
        if not len(self.tags):
            maxtag = b"T000"
        else:
            maxtag = max(self.tags)
        newtag = b"T%03d" % (int(maxtag[1:]) + 1)
        return newtag

    def request(self, command):
        if not (self.session and command.session) and command.command not in (b'AUTH', b'PING', b'ENCRYPT'):
            raise AniDBMustAuthError("You must be authed to execute commands besides AUTH and PING")
        command.started = time()
        self._cmd_queue(command)
        self._send(command)



from __future__ import absolute_import, division, print_function

from trakt.core.configuration import ConfigurationManager
from trakt.core.emitter import Emitter
from trakt.core.http import HttpClient
from trakt.interfaces import construct_map
from trakt.interfaces.base import InterfaceProxy
from trakt.version import __version__


class TraktClient(Emitter):
    base_url = 'https://api.trakt.tv'
    version = __version__

    __interfaces = None

    def __init__(self, adapter_kwargs=None):
        # Set parameter defaults
        if adapter_kwargs is None:
            adapter_kwargs = {}

        adapter_kwargs.setdefault('max_retries', 3)

        # Construct
        self.configuration = ConfigurationManager()
        self.http = HttpClient(self, adapter_kwargs)

        self.__interfaces = construct_map(self)

        self._site_url = None

    @property
    def site_url(self):
        if self._site_url is not None:
            return self._site_url

        url = self.base_url

        schema_end = url.find('://') + 3
        domain_start = url.find('.', schema_end) + 1

        return url[0:schema_end] + url[domain_start:]

    @site_url.setter
    def site_url(self, value):
        self._site_url = value

    def __getitem__(self, path):
        parts = path.strip('/').split('/')

        cur = self.__interfaces
        parameters = []

        while parts and type(cur) is dict:
            key = parts.pop(0)

            if key not in cur:
                if '*' in cur:
                    if key != '*':
                        parameters.append(key)

                    cur = cur['*']
                    continue

                return None

            cur = cur[key]

        if type(cur) is dict:
            cur = cur.get(None)

        if parts:
            parameters.extend(parts)

        if parameters:
            return InterfaceProxy(cur, parameters)

        return cur



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################



from upnpclient import const, errors, marshal, soap, ssdp, upnp, util  # noqa: F401
from .ssdp import discover
from .upnp import (
    Device, Action, Service, UPNPError, InvalidActionException, ValidationError, UnexpectedResponse)

__all__ = [
    "Device", "Action", "Service", "UPNPError", "InvalidActionException", "ValidationError",
    "discover", "UnexpectedResponse"
]



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################


from base64 import b64encode

import sickrage
from sickrage.clients import TorrentClient
from sickrage.core.tv.show.helpers import find_show


class DelugeAPI(TorrentClient):
    def __init__(self, host=None, username=None, password=None):

        super(DelugeAPI, self).__init__('Deluge', host, username, password)

        self.url = self.host + 'json'
        self.session.headers.update({'Content-type': "application/json"})

    def _get_auth(self):
        post_data = {"method": "auth.login",
                     "params": [self.password],
                     "id": 1}

        self.response = self.session.post(self.url, json=post_data, verify=bool(sickrage.app.config.torrent_verify_cert))
        if not self.response or not self.response.content:
            return None

        try:
            data = self.response.json()
        except ValueError:
            return None

        self.auth = data.get("result")

        post_data = {
            "method": "web.connected",
            "params": [],
            "id": 10
        }

        self.response = self.session.post(self.url, json=post_data, verify=bool(sickrage.app.config.torrent_verify_cert))
        if not self.response or not self.response.content:
            return None

        try:
            connected = self.response.json()
        except ValueError:
            return None

        if not connected.get('result'):
            post_data = {
                "method": "web.get_hosts",
                "params": [],
                "id": 11
            }

            self.response = self.session.post(self.url, json=post_data, verify=bool(sickrage.app.config.torrent_verify_cert))
            if not self.response or not self.response.content:
                return None

            try:
                hosts = self.response.json()
            except ValueError:
                return None

            if not hosts.get('result'):
                sickrage.app.log.warning(self.name + ': WebUI does not contain daemons')
                return None

            post_data = {
                "method": "web.connect",
                "params": [hosts[0][0]],
                "id": 11
            }

            self.response = self.session.post(self.url, json=post_data, verify=bool(sickrage.app.config.torrent_verify_cert))
            if not self.response:
                return None

            post_data = {
                "method": "web.connected",
                "params": [],
                "id": 10
            }

            self.response = self.session.post(self.url, json=post_data, verify=bool(sickrage.app.config.torrent_verify_cert))
            if not self.response or not self.response.content:
                return None

            try:
                connected = self.response.json()
            except ValueError:
                return None

            if not connected.get('result'):
                sickrage.app.log.warning(self.name + ': WebUI could not connect to daemon')
                return None

        return self.auth

    def _add_torrent_uri(self, result):
        post_data = {"method": "core.add_torrent_magnet",
                     "params": [result.url, {}],
                     "id": 2}

        self._request(method='post', json=post_data)

        result.hash = self.response.json()['result']

        return self.response.json()['result']

    def _add_torrent_file(self, result):
        post_data = {"method": "core.add_torrent_file",
                     "params": [result.name + '.torrent', b64encode(result.content), {}],
                     "id": 2}

        self._request(method='post', json=post_data)

        result.hash = self.response.json()['result']

        return self.response.json()['result']

    def _set_torrent_label(self, result):
        label = sickrage.app.config.torrent_label

        tv_show = find_show(result.show_id)

        if tv_show.is_anime:
            label = sickrage.app.config.torrent_label_anime

        if ' ' in label:
            sickrage.app.log.warning(self.name + ': Invalid label. Label must not contain a space')
            return False

        if label:
            # check if label already exists and create it if not
            post_data = {"method": 'label.get_labels',
                         "params": [],
                         "id": 3}

            self._request(method='post', json=post_data)
            labels = self.response.json()['result']

            if labels is not None:
                if label not in labels:
                    sickrage.app.log.debug(self.name + ': ' + label + " label does not exist in Deluge we must add it")
                    post_data = {"method": 'label.add',
                                 "params": [label],
                                 "id": 4}

                    self._request(method='post', json=post_data)
                    sickrage.app.log.debug(self.name + ': ' + label + " label added to Deluge")

                # add label to torrent
                post_data = {"method": 'label.set_torrent',
                             "params": [result.hash, label],
                             "id": 5}

                self._request(method='post', json=post_data)
                sickrage.app.log.debug(self.name + ': ' + label + " label added to torrent")
            else:
                sickrage.app.log.debug(self.name + ': ' + "label plugin not detected")
                return False

        return not self.response.json()['error']

    def _set_torrent_ratio(self, result):
        ratio = None
        if result.ratio:
            ratio = result.ratio

        if ratio:
            post_data = {"method": "core.set_torrent_stop_at_ratio",
                         "params": [result.hash, True],
                         "id": 5}

            self._request(method='post', json=post_data)

            post_data = {"method": "core.set_torrent_stop_ratio",
                         "params": [result.hash, float(ratio)],
                         "id": 6}

            self._request(method='post', json=post_data)

            return not self.response.json()['error']

        return True

    def _set_torrent_path(self, result):
        if sickrage.app.config.torrent_path:
            post_data = {"method": "core.set_torrent_move_completed",
                         "params": [result.hash, True],
                         "id": 7}

            self._request(method='post', json=post_data)

            post_data = {"method": "core.set_torrent_move_completed_path",
                         "params": [result.hash, sickrage.app.config.torrent_path],
                         "id": 8}

            self._request(method='post', json=post_data)

            return not self.response.json()['error']

        return True

    def _set_torrent_pause(self, result):
        if sickrage.app.config.torrent_paused:
            post_data = {"method": "core.pause_torrent",
                         "params": [[result.hash]],
                         "id": 9}

            self._request(method='post', json=post_data)

            return not self.response.json()['error']

        return True



# URL: https://sickrage.ca
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.



import datetime
from xml.etree.ElementTree import Element, ElementTree, SubElement

import sickrage
from sickrage.core.common import dateFormat
from sickrage.core.helpers import indent_xml
from sickrage.indexers import IndexerApi
from sickrage.indexers.exceptions import indexer_episodenotfound, \
    indexer_seasonnotfound
from sickrage.metadata import GenericMetadata


class KODI_12PlusMetadata(GenericMetadata):
    """
    Metadata generation class for KODI 12+.

    The following file structure is used:

    show_root/tvshow.nfo                    (show metadata)
    show_root/fanart.jpg                    (fanart)
    show_root/poster.jpg                    (poster)
    show_root/banner.jpg                    (banner)
    show_root/Season ##/filename.ext        (*)
    show_root/Season ##/filename.nfo        (episode metadata)
    show_root/Season ##/filename-thumb.jpg  (episode thumb)
    show_root/season##-poster.jpg           (season posters)
    show_root/season##-banner.jpg           (season banners)
    show_root/season-all-poster.jpg         (season all poster)
    show_root/season-all-banner.jpg         (season all banner)
    """

    def __init__(self,
                 show_metadata=False,
                 episode_metadata=False,
                 fanart=False,
                 poster=False,
                 banner=False,
                 episode_thumbnails=False,
                 season_posters=False,
                 season_banners=False,
                 season_all_poster=False,
                 season_all_banner=False):

        GenericMetadata.__init__(self,
                                 show_metadata,
                                 episode_metadata,
                                 fanart,
                                 poster,
                                 banner,
                                 episode_thumbnails,
                                 season_posters,
                                 season_banners,
                                 season_all_poster,
                                 season_all_banner)

        self.name = 'KODI 12+'

        self.poster_name = "poster.jpg"
        self.season_all_poster_name = "season-all-poster.jpg"

        # web-ui metadata template
        self.eg_show_metadata = "tvshow.nfo"
        self.eg_episode_metadata = "Season##\\<i>filename</i>.nfo"
        self.eg_fanart = "fanart.jpg"
        self.eg_poster = "poster.jpg"
        self.eg_banner = "banner.jpg"
        self.eg_episode_thumbnails = "Season##\\<i>filename</i>-thumb.jpg"
        self.eg_season_posters = "season##-poster.jpg"
        self.eg_season_banners = "season##-banner.jpg"
        self.eg_season_all_poster = "season-all-poster.jpg"
        self.eg_season_all_banner = "season-all-banner.jpg"

    def _show_data(self, show_obj):
        """
        Creates an elementTree XML structure for an KODI-style tvshow.nfo and
        returns the resulting data object.

        show_obj: a TVShow instance to create the NFO for
        """

        tv_node = Element("tvshow")

        show_ID = show_obj.indexer_id

        indexer_lang = show_obj.lang or sickrage.app.config.indexer_default_language

        lINDEXER_API_PARMS = IndexerApi(show_obj.indexer).api_params.copy()
        lINDEXER_API_PARMS['language'] = indexer_lang

        if show_obj.dvdorder != 0:
            lINDEXER_API_PARMS['dvdorder'] = True

        t = IndexerApi(show_obj.indexer).indexer(**lINDEXER_API_PARMS)
        indexer_data = t[int(show_ID)]
        if not indexer_data:
            return False

        # check for title and id
        if not (getattr(indexer_data, 'seriesname', None) and getattr(indexer_data, 'id', None)):
            sickrage.app.log.info("Incomplete info for show with id " + str(show_ID) + " on " + IndexerApi(show_obj.indexer).name + ", skipping it")
            return False

        title = SubElement(tv_node, "title")
        title.text = indexer_data["seriesname"]

        if getattr(indexer_data, 'rating', None):
            rating = SubElement(tv_node, "rating")
            rating.text = indexer_data["rating"]

        if getattr(indexer_data, 'firstaired', None):
            try:
                year_text = str(datetime.datetime.strptime(indexer_data["firstaired"], dateFormat).year)
                if year_text:
                    year = SubElement(tv_node, "year")
                    year.text = year_text
            except Exception:
                pass

        if getattr(indexer_data, 'overview', None):
            plot = SubElement(tv_node, "plot")
            plot.text = indexer_data["overview"]

        # if getattr(indexer_data, 'id', None):
        #    episodeguide = SubElement(tv_node, "episodeguide")
        #    episodeguideurl = SubElement(episodeguide, "url")
        #    episodeguideurl.text = IndexerApi(show_obj.indexer).config['base_url'] + str(
        #        indexer_data["id"]) + '/all/en.zip'

        if getattr(indexer_data, 'contentrating', None):
            mpaa = SubElement(tv_node, "mpaa")
            mpaa.text = indexer_data["contentrating"]

        if getattr(indexer_data, 'id', None):
            indexer_id = SubElement(tv_node, "id")
            indexer_id.text = str(indexer_data["id"])

        if getattr(indexer_data, 'genre', None) and isinstance(indexer_data["genre"], str):
            genre = SubElement(tv_node, "genre")
            genre.text = " / ".join(x.strip() for x in indexer_data["genre"].split('|') if x.strip())

        if getattr(indexer_data, 'firstaired', None):
            premiered = SubElement(tv_node, "premiered")
            premiered.text = indexer_data["firstaired"]

        if getattr(indexer_data, 'network', None):
            studio = SubElement(tv_node, "studio")
            studio.text = indexer_data["network"].strip()

        for actor in t.actors(int(show_ID)):
            cur_actor = SubElement(tv_node, "actor")

            if 'name' in actor and actor['name'].strip():
                cur_actor_name = SubElement(cur_actor, "name")
                cur_actor_name.text = actor['name'].strip()
            else:
                continue

            if 'role' in actor and actor['role'].strip():
                cur_actor_role = SubElement(cur_actor, "role")
                cur_actor_role.text = actor['role'].strip()

            if 'image' in actor and actor['image'].strip():
                cur_actor_thumb = SubElement(cur_actor, "thumb")
                cur_actor_thumb.text = actor['image'].strip()

        # Make it purdy
        indent_xml(tv_node)

        data = ElementTree(tv_node)

        return data

    def _ep_data(self, ep_obj):
        """
        Creates an elementTree XML structure for an KODI-style episode.nfo and
        returns the resulting data object.
            show_obj: a TVEpisode instance to create the NFO for
        """

        eps_to_write = [ep_obj] + ep_obj.related_episodes

        indexer_lang = ep_obj.show.lang or sickrage.app.config.indexer_default_language

        lINDEXER_API_PARMS = IndexerApi(ep_obj.show.indexer).api_params.copy()
        lINDEXER_API_PARMS['language'] = indexer_lang

        if ep_obj.show.dvdorder != 0:
            lINDEXER_API_PARMS['dvdorder'] = True

        t = IndexerApi(ep_obj.show.indexer).indexer(**lINDEXER_API_PARMS)
        indexer_data = t[ep_obj.show.indexer_id]
        if not indexer_data:
            return False

        if len(eps_to_write) > 1:
            rootNode = Element("kodimultiepisode")
        else:
            rootNode = Element("episodedetails")

        # write an NFO containing info for all matching episodes
        for curEpToWrite in eps_to_write:

            try:
                myEp = indexer_data[curEpToWrite.season][curEpToWrite.episode]
            except (indexer_episodenotfound, indexer_seasonnotfound):
                sickrage.app.log.info(
                    "Unable to find episode %dx%d on %s... has it been removed? Should I delete from db?" %
                    (curEpToWrite.season, curEpToWrite.episode, IndexerApi(ep_obj.show.indexer).name))
                return None

            if not getattr(myEp, 'firstaired', None):
                myEp["firstaired"] = str(datetime.date.min)

            if not getattr(myEp, 'episodename', None):
                sickrage.app.log.debug("Not generating nfo because the ep has no title")
                return None

            sickrage.app.log.debug(
                "Creating metadata for episode " + str(ep_obj.season) + "x" + str(ep_obj.episode))

            if len(eps_to_write) > 1:
                episode = SubElement(rootNode, "episodedetails")
            else:
                episode = rootNode

            if getattr(myEp, 'episodename', None):
                title = SubElement(episode, "title")
                title.text = myEp['episodename']

            if getattr(indexer_data, 'seriesname', None):
                showtitle = SubElement(episode, "showtitle")
                showtitle.text = indexer_data['seriesname']

            season = SubElement(episode, "season")
            season.text = str(curEpToWrite.season)

            episodenum = SubElement(episode, "episode")
            episodenum.text = str(curEpToWrite.episode)

            uniqueid = SubElement(episode, "uniqueid")
            uniqueid.text = str(curEpToWrite.indexer_id)

            if curEpToWrite.airdate > datetime.date.min:
                aired = SubElement(episode, "aired")
                aired.text = str(curEpToWrite.airdate)

            if getattr(myEp, 'overview', None):
                plot = SubElement(episode, "plot")
                plot.text = myEp['overview']

            if curEpToWrite.season and getattr(indexer_data, 'runtime', None):
                runtime = SubElement(episode, "runtime")
                runtime.text = indexer_data["runtime"]

            if getattr(myEp, 'airsbefore_season', None):
                displayseason = SubElement(episode, "displayseason")
                displayseason.text = myEp['airsbefore_season']

            if getattr(myEp, 'airsbefore_episode', None):
                displayepisode = SubElement(episode, "displayepisode")
                displayepisode.text = myEp['airsbefore_episode']

            if getattr(myEp, 'filename', None):
                thumb = SubElement(episode, "thumb")
                thumb.text = myEp['filename'].strip()

            # watched = SubElement(episode, "watched")
            # watched.text = 'false'

            if getattr(myEp, 'writer', None):
                ep_credits = SubElement(episode, "credits")
                ep_credits.text = myEp['writer'].strip()

            if getattr(myEp, 'director', None):
                director = SubElement(episode, "director")
                director.text = myEp['director'].strip()

            if getattr(myEp, 'rating', None):
                rating = SubElement(episode, "rating")
                rating.text = myEp['rating']

            if getattr(myEp, 'gueststars', None) and isinstance(myEp['gueststars'], str):
                for actor in (x.strip() for x in myEp['gueststars'].split('|') if x.strip()):
                    cur_actor = SubElement(episode, "actor")
                    cur_actor_name = SubElement(cur_actor, "name")
                    cur_actor_name.text = actor

            for actor in t.actors(int(ep_obj.show.indexer_id)):
                cur_actor = SubElement(episode, "actor")

                if 'name' in actor and actor['name'].strip():
                    cur_actor_name = SubElement(cur_actor, "name")
                    cur_actor_name.text = actor['name'].strip()
                else:
                    continue

                if 'role' in actor and actor['role'].strip():
                    cur_actor_role = SubElement(cur_actor, "role")
                    cur_actor_role.text = actor['role'].strip()

                if 'image' in actor and actor['image'].strip():
                    cur_actor_thumb = SubElement(cur_actor, "thumb")
                    cur_actor_thumb.text = actor['image'].strip()

        # Make it purdy
        indent_xml(rootNode)
        data = ElementTree(rootNode)
        return data



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################


import datetime
import functools
import time

import feedparser
from sqlalchemy import orm
from sqlalchemy.exc import IntegrityError

import sickrage
from sickrage.core.common import Quality
from sickrage.core.databases.cache import CacheDB
from sickrage.core.exceptions import AuthException
from sickrage.core.helpers import show_names, try_int
from sickrage.core.nameparser import InvalidNameException, NameParser, InvalidShowException
from sickrage.core.tv.show.helpers import find_show
from sickrage.core.websession import WebSession


class TVCache(object):
    def __init__(self, provider, **kwargs):
        self.provider = provider
        self.providerID = self.provider.id
        self.min_time = kwargs.pop('min_time', 10)
        self.search_strings = kwargs.pop('search_strings', dict(RSS=['']))

    def clear(self):
        session = sickrage.app.cache_db.session()
        if self.shouldClearCache():
            session.query(CacheDB.Provider).filter_by(provider=self.providerID).delete()
            session.commit()

    def _get_title_and_url(self, item):
        return self.provider._get_title_and_url(item)

    def _get_result_stats(self, item):
        return self.provider._get_result_stats(item)

    def _get_size(self, item):
        return self.provider._get_size(item)

    def _get_rss_data(self):
        if self.search_strings:
            return {'entries': self.provider.search(self.search_strings)}

    def _check_auth(self, data):
        return True

    def check_item(self, title, url):
        return True

    def update(self, force=False):
        # check if we should update
        if self.should_update() or force:
            try:
                data = self._get_rss_data()
                if not self._check_auth(data):
                    return False

                # clear cache
                self.clear()

                # set updated
                self.last_update = datetime.datetime.today()

                [self._parseItem(item) for item in data['entries']]

                sickrage.app.log.debug("Updated RSS cache")
            except AuthException as e:
                sickrage.app.log.warning("Authentication error: {}".format(e))
                return False
            except Exception as e:
                sickrage.app.log.debug("Error while searching {}, skipping: {}".format(self.provider.name, repr(e)))
                return False

        return True

    def get_rss_feed(self, url, params=None):
        try:
            if self.provider.login():
                resp = WebSession().get(url, timeout=30, params=params)
                if resp:
                    return feedparser.parse(resp.text)
        except Exception as e:
            sickrage.app.log.debug("RSS Error: {}".format(e))

        return feedparser.FeedParserDict()

    def _translateTitle(self, title):
        return '' + title.replace(' ', '.')

    def _translateLinkURL(self, url):
        return url.replace('&amp;', '&')

    def _parseItem(self, item):
        title, url = self._get_title_and_url(item)
        seeders, leechers = self._get_result_stats(item)
        size = self._get_size(item)

        self.check_item(title, url)

        if title and url:
            self.add_cache_entry(self._translateTitle(title), self._translateLinkURL(url), seeders, leechers, size)
        else:
            sickrage.app.log.debug(
                "The data returned from the " + self.provider.name + " feed is incomplete, this result is unusable")

    @property
    def last_update(self):
        session = sickrage.app.cache_db.session()

        try:
            dbData = session.query(CacheDB.LastUpdate).filter_by(provider=self.providerID).one()
            lastTime = int(dbData.time)
            if lastTime > int(time.mktime(datetime.datetime.today().timetuple())):
                lastTime = 0
        except orm.exc.NoResultFound:
            lastTime = 0

        return datetime.datetime.fromtimestamp(lastTime)

    @last_update.setter
    def last_update(self, toDate):
        session = sickrage.app.cache_db.session()

        try:
            dbData = session.query(CacheDB.LastUpdate).filter_by(provider=self.providerID).one()
            dbData.time = int(time.mktime(toDate.timetuple()))
        except orm.exc.NoResultFound:
            session.add(CacheDB.LastUpdate(**{
                'provider': self.providerID,
                'time': int(time.mktime(toDate.timetuple()))
            }))
        finally:
            session.commit()

    @property
    def last_search(self):
        session = sickrage.app.cache_db.session()

        try:
            dbData = session.query(CacheDB.LastSearch).filter_by(provider=self.providerID).one()
            lastTime = int(dbData.time)
            if lastTime > int(time.mktime(datetime.datetime.today().timetuple())):
                lastTime = 0
        except orm.exc.NoResultFound:
            lastTime = 0

        return datetime.datetime.fromtimestamp(lastTime)

    @last_search.setter
    def last_search(self, toDate):
        session = sickrage.app.cache_db.session()

        try:
            dbData = session.query(CacheDB.LastSearch).filter_by(provider=self.providerID).one()
            dbData.time = int(time.mktime(toDate.timetuple()))
        except orm.exc.NoResultFound:
            session.add(CacheDB.LastSearch(**{
                'provider': self.providerID,
                'time': int(time.mktime(toDate.timetuple()))
            }))
        finally:
            session.commit()

    def should_update(self):
        # if we've updated recently then skip the update
        if datetime.datetime.today() - self.last_update < datetime.timedelta(minutes=self.min_time):
            return False
        return True

    def shouldClearCache(self):
        # if daily search hasn't used our previous results yet then don't clear the cache
        if self.last_update > self.last_search:
            return False
        return True

    def add_cache_entry(self, name, url, seeders, leechers, size):
        session = sickrage.app.cache_db.session()

        # check for existing entry in cache
        if session.query(CacheDB.Provider).filter_by(url=url).count():
            return

        try:
            # parse release name
            parse_result = NameParser(validate_show=True).parse(name)
            if parse_result.series_name and parse_result.quality != Quality.UNKNOWN:
                season = parse_result.season_number if parse_result.season_number else 1
                episodes = parse_result.episode_numbers

                if season and episodes:
                    # store episodes as a seperated string
                    episode_text = "|" + "|".join(map(str, episodes)) + "|"

                    # get quality of release
                    quality = parse_result.quality

                    # get release group
                    release_group = parse_result.release_group

                    # get version
                    version = parse_result.version

                    dbData = {
                        'provider': self.providerID,
                        'name': name,
                        'season': season,
                        'episodes': episode_text,
                        'series_id': parse_result.indexer_id,
                        'url': url,
                        'time': int(time.mktime(datetime.datetime.today().timetuple())),
                        'quality': quality,
                        'release_group': release_group,
                        'version': version,
                        'seeders': try_int(seeders),
                        'leechers': try_int(leechers),
                        'size': try_int(size, -1)
                    }

                    # add to internal database
                    try:
                        session.add(CacheDB.Provider(**dbData))
                        session.commit()
                        sickrage.app.log.debug("SEARCH RESULT:[{}] ADDED TO CACHE!".format(name))
                    except IntegrityError:
                        pass

                    # add to external provider cache database
                    if sickrage.app.config.enable_sickrage_api and not self.provider.private and self.provider.type in ['nzb', 'torrent']:
                        try:
                            sickrage.app.api.provider_cache.add(data=dbData)
                        except Exception as e:
                            pass
        except (InvalidShowException, InvalidNameException):
            pass

    def search_cache(self, show_id, season, episode, manualSearch=False, downCurQuality=False):
        cache_results = {}
        dbData = []

        # get data from external database
        if sickrage.app.config.enable_sickrage_api and not self.provider.private:
            resp = sickrage.app.api.provider_cache.get(self.providerID, show_id, season, episode)
            if resp and 'data' in resp:
                dbData += resp['data']

        # get data from internal database
        session = sickrage.app.cache_db.session()
        dbData += [x.as_dict() for x in
                   session.query(CacheDB.Provider).filter_by(provider=self.providerID,
                                                             series_id=show_id,
                                                             season=season).filter(CacheDB.Provider.episodes.contains("|{}|".format(episode)))]

        for curResult in dbData:
            show_object = find_show(int(curResult["series_id"]))
            if not show_object:
                continue

            result = self.provider.get_result()

            # ignored/required words, and non-tv junk
            if not show_names.filter_bad_releases(curResult["name"]):
                continue

            # get the show object, or if it's not one of our shows then ignore it
            result.show_id = int(curResult["series_id"])

            # skip if provider is anime only and show is not anime
            if self.provider.anime_only and not show_object.is_anime:
                sickrage.app.log.debug("" + str(show_object.name) + " is not an anime, skiping")
                continue

            # get season and ep data (ignoring multi-eps for now)
            curSeason = int(curResult["season"])
            if curSeason == -1:
                continue

            result.season = curSeason
            result.episodes = [int(curEp) for curEp in filter(None, curResult["episodes"].split("|"))]

            result.quality = int(curResult["quality"])
            result.release_group = curResult["release_group"]
            result.version = curResult["version"]

            # make sure we want the episode
            wantEp = False
            for result_episode in result.episodes:
                if show_object.want_episode(result.season, result_episode, result.quality, manualSearch, downCurQuality):
                    wantEp = True

            if not wantEp:
                sickrage.app.log.info("Skipping " + curResult["name"] + " because we don't want an episode that's " + Quality.qualityStrings[result.quality])
                continue

            # build a result object
            result.name = curResult["name"]
            result.url = curResult["url"]

            sickrage.app.log.info("Found cached {} result {}".format(result.type, result.name))

            result.seeders = curResult.get("seeders", -1)
            result.leechers = curResult.get("leechers", -1)
            result.size = curResult.get("size", -1)
            result.content = None

            # add it to the list
            if episode not in cache_results:
                cache_results[int(episode)] = [result]
            else:
                cache_results[int(episode)] += [result]

        # datetime stamp this search so cache gets cleared
        self.last_search = datetime.datetime.today()

        return cache_results



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################
import functools
import threading

from sqlalchemy import orm

import sickrage
from sickrage.core.databases.cache import CacheDB


class Announcement(object):
    """
    Represents an announcement.
    """

    def __init__(self, title, description, image, date, ahash):
        self.title = title
        self.description = description
        self.image = image
        self.date = date
        self.ahash = ahash

    @property
    def seen(self):
        session = sickrage.app.cache_db.session()
        try:
            announcement = session.query(CacheDB.Announcements).filter_by(hash=self.ahash).one()
            return True if announcement.seen else False
        except orm.exc.NoResultFound:
            pass

    @seen.setter
    def seen(self, value):
        session = sickrage.app.cache_db.session()
        try:
            announcement = session.query(CacheDB.Announcements).filter_by(hash=self.ahash).one()
            announcement.seen = value
        except orm.exc.NoResultFound:
            pass


class Announcements(object):
    """
    Keeps a static list of (announcement) UIErrors to be displayed on the UI and allows
    the list to be cleared.
    """

    def __init__(self):
        self.name = "ANNOUNCEMENTS"
        self.running = False
        self._announcements = {}

    def task(self, force=False):
        if self.running and not force:
            return

        try:
            self.running = True

            threading.currentThread().setName(self.name)

            resp = sickrage.app.api.announcement.get_announcements()
            if resp and 'data' in resp:
                for announcement in resp['data']:
                    if announcement['enabled']:
                        self.add(announcement['hash'], announcement['title'], announcement['description'], announcement['image'], announcement['date'])
                    else:
                        self.clear(announcement['hash'])
        finally:
            self.running = False

    def add(self, ahash, title, description, image, date):
        session = sickrage.app.cache_db.session()
        self._announcements[ahash] = Announcement(title, description, image, date, ahash)
        if not session.query(CacheDB.Announcements).filter_by(hash=ahash).count():
            sickrage.app.log.debug('Adding new announcement to Web-UI')
            session.add(CacheDB.Announcements(**{'hash': ahash}))
            session.commit()

    def clear(self, ahash=None):
        session = sickrage.app.cache_db.session()

        if not ahash:
            self._announcements.clear()
            session.query(CacheDB.Announcements).delete()
            session.commit()
        else:
            if ahash in self._announcements:
                del self._announcements[ahash]
            session.query(CacheDB.Announcements).filter_by(hash=ahash).delete()
            session.commit()

    def get_all(self):
        return sorted(self._announcements.values(), key=lambda k: k.date)

    def get(self, ahash):
        return self._announcements.get(ahash)

    def count(self):
        session = sickrage.app.cache_db.session()
        return session.query(CacheDB.Announcements).filter(CacheDB.Announcements.seen is False).count()



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
# Git: https://git.sickrage.ca/SiCKRAGE/sickrage
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
from urllib.parse import urljoin

import sickrage
from sickrage.core.caches.tv_cache import TVCache
from sickrage.core.helpers import try_int, convert_size, validate_url
from sickrage.core.tv.show.helpers import find_show
from sickrage.providers import TorrentProvider


class BitCannonProvider(TorrentProvider):
    def __init__(self):
        super(BitCannonProvider, self).__init__("BitCannon", 'http://localhost:3000', False)
        self.api_key = None

        self.minseed = None
        self.minleech = None

        self.custom_url = ""

        self._urls.update({
            'search': '{base_url}/api/search'.format(**self._urls)
        })

        self.cache = TVCache(self, search_strings={'RSS': ['tv', 'anime']})

    def search(self, search_strings, age=0, show_id=None, season=None, episode=None, **kwargs):
        results = []

        search_url = self.urls["search"]
        if self.custom_url:
            if not validate_url(self.custom_url):
                sickrage.app.log.warning("Invalid custom url: {0}".format(self.custom_url))
                return results
            search_url = urljoin(self.custom_url, search_url.split(self.urls['base_url'])[1])

        show_object = find_show(show_id)

        # Search Params
        search_params = {
            'category': ("tv", "anime")[bool(show_object.anime)],
            'apiKey': self.api_key,
        }

        for mode in search_strings:
            sickrage.app.log.debug('Search mode: {}'.format(mode))
            for search_string in search_strings[mode]:
                search_params['q'] = search_string
                if mode != 'RSS':
                    sickrage.app.log.debug('Search string: {}'.format(search_string))

                resp = self.session.get(search_url, params=search_params)
                if not resp or not resp.content:
                    sickrage.app.log.debug('No data returned from provider')
                    continue

                try:
                    data = resp.json()
                except ValueError:
                    sickrage.app.log.debug('No data returned from provider')
                    continue

                results += self.parse(data, mode)

        return results

    def parse(self, data, mode, **kwargs):
        """
        Parse search results from data
        :param data: response data
        :param mode: search mode
        :return: search results
        """

        results = []

        torrent_rows = data.pop('torrents', {})

        if not self._check_auth_from_data(data):
            return results

        # Skip column headers
        for row in torrent_rows:
            try:
                title = row.pop('title', '')
                info_hash = row.pop('infoHash', '')
                download_url = 'magnet:?xt=urn:btih:' + info_hash
                if not all([title, download_url, info_hash]):
                    continue

                swarm = row.pop('swarm', {})
                seeders = try_int(swarm.pop('seeders'))
                leechers = try_int(swarm.pop('leechers'))

                size = convert_size(row.pop('size', -1), -1)

                results += [
                    {'title': title, 'link': download_url, 'size': size, 'seeders': seeders, 'leechers': leechers}
                ]

                if mode != 'RSS':
                    sickrage.app.log.debug('Found result: {}'.format(title))
            except Exception:
                sickrage.app.log.error('Failed parsing provider')

        return results

    @staticmethod
    def _check_auth_from_data(data):
        if not all([isinstance(data, dict),
                    data.pop('status', 200) != 401,
                    data.pop('message', '') != 'Invalid API key']):
            sickrage.app.log.warning('Invalid api key. Check your settings')
            return False

        return True



from sqlalchemy import *
from migrate import *


def upgrade(migrate_engine):
    # Upgrade operations go here. Don't create your own engine; bind
    # migrate_engine to your metadata
    pass


def downgrade(migrate_engine):
    # Operations to reverse the above upgrade go here.
    pass



from __future__ import absolute_import, division, print_function

from trakt.mapper.core.base import Mapper


class CommentMapper(Mapper):
    @classmethod
    def comment(cls, client, item, **kwargs):
        if 'comment' in item:
            i_comment = item['comment']
        else:
            i_comment = item

        # Retrieve item keys
        pk, keys = cls.get_ids('comment', i_comment)

        if pk is None:
            return None

        # Create object
        comment = cls.construct(client, 'comment', i_comment, keys, **kwargs)

        # Update with root info
        if 'comment' in item:
            comment._update(item)

        return comment



from __future__ import absolute_import, division, print_function

from threading import Lock


class KeyLock(object):
    def __init__(self, lock=Lock):
        self._lock = lock

        self._create_lock = Lock()
        self._locks = {}

    def __getitem__(self, key):
        # Return lock if it's available
        if key in self._locks:
            return self._locks[key]

        # Create new lock for `key` (synchronized)
        with self._create_lock:
            if key not in self._locks:
                self._locks[key] = self._lock()

            return self._locks[key]



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################

from sickrage.core.webserver.handlers.base import BaseHandler


@Route('/googleDrive(/?.*)')
class GoogleDriveHandler(BaseHandler):
    def __init__(self, *args, **kwargs):
        super(GoogleDriveHandler, self).__init__(*args, **kwargs)

    def getProgress(self):
        return google_drive.GoogleDrive.get_progress()

    def syncRemote(self):
        self._genericMessage(_("Google Drive Sync"), _("Syncing app data to Google Drive"))
        google_drive.GoogleDrive().sync_remote()

    def syncLocal(self):
        self._genericMessage(_("Google Drive Sync"), _("Syncing app data from Google Drive"))
        google_drive.GoogleDrive().sync_local()


# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################
import json
import os
from json import JSONDecodeError

from sqlalchemy import *

import sickrage


def upgrade(migrate_engine):
    meta = MetaData(bind=migrate_engine)
    oauth2_token = Table('oauth2_token', meta, autoload=True)
    token_file = os.path.abspath(os.path.join(sickrage.app.data_dir, 'token.json'))

    if os.path.exists(token_file):
        with open(token_file, 'r') as fd:
            try:
                token = json.load(fd)
                migrate_engine.execute(oauth2_token.insert().values(
                    access_token=token['access_token'],
                    refresh_token=token['refresh_token'],
                    expires_in=token['expires_in'],
                    expires_at=token['expires_at'],
                    scope=' '.join(token['scope']) if isinstance(token['scope'], list) else token['scope']
                ))
            except JSONDecodeError:
                pass

        os.remove(token_file)


def downgrade(migrate_engine):
    # Operations to reverse the above upgrade go here.
    pass



from __future__ import absolute_import, division, print_function

from trakt.core.helpers import from_iso8601_datetime


class Rating(object):
    def __init__(self, client, value=None, timestamp=None):
        self._client = client

        self.value = value
        """
        :type: :class:`~python:int`

        Rating value (0 - 10)
        """

        self.timestamp = timestamp
        """
        :type: :class:`~python:datetime.datetime`

        Rating timestamp
        """

    @classmethod
    def _construct(cls, client, info):
        if not info or 'rating' not in info:
            return

        r = cls(client)
        r.value = info.get('rating')
        r.timestamp = from_iso8601_datetime(info.get('rated_at'))
        return r

    def __getstate__(self):
        state = self.__dict__

        if hasattr(self, '_client'):
            del state['_client']

        return state

    def __eq__(self, other):
        if not isinstance(other, Rating):
            return NotImplemented

        return self.value == other.value and self.timestamp == other.timestamp

    def __repr__(self):
        return '<Rating %s/10 (%s)>' % (self.value, self.timestamp)

    def __str__(self):
        return self.__repr__()



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################
import json
from abc import ABC

import sickrage
from sickrage.core.webserver.handlers.base import BaseHandler
from sickrage.libs.trakt.interfaces.base import authenticated


class AnnouncementsHandler(BaseHandler, ABC):
    @authenticated
    async def get(self, *args, **kwargs):
        await self.run_in_executor(self.handle_get)

    def handle_get(self):
        return self.render('announcements.mako',
                           announcements=sickrage.app.announcements.get_all(),
                           title=_('Announcements'),
                           header=_('Announcements'),
                           topmenu='announcements',
                           controller='root',
                           action='announcements')


class MarkAnnouncementSeenHandler(BaseHandler, ABC):
    @authenticated
    async def post(self, *args, **kwargs):
        ahash = self.get_argument('ahash')

        announcement = sickrage.app.announcements.get(ahash)
        if announcement:
            announcement.seen = True

        return self.write(json.dumps({'success': True}))


class AnnouncementCountHandler(BaseHandler, ABC):
    @authenticated
    async def get(self, *args, **kwargs):
        await self.run_in_executor(self.handle_get)

    def handle_get(self):
        return self.write(json.dumps({'count': sickrage.app.announcements.count()}))



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################
from urllib.parse import urljoin, urlencode
from xml.dom.minidom import parseString

from requests import request

import sickrage
from sickrage.notifiers import Notifiers


class NMA_Notifier(Notifiers):
    def __init__(self):
        super(NMA_Notifier, self).__init__()
        self.name = 'nma'
        self.url = 'https://www.notifymyandroid.com'
        self._developerkey = None
        self._apikey = None

    def uniq_preserve(self, seq):  # Dave Kirby
        # Order preserving
        seen = set()
        return [x for x in seq if x not in seen and not seen.add(x)]

    def uniq(self, seq):
        # Not order preserving
        return list({}.fromkeys(seq).keys())

    def addkey(self, key):
        """Add a key (register ?)"""
        self._apikey = self.uniq(key)
        if type(key) == str:
            if not key in self._apikey:
                self._apikey.append(key)
        elif type(key) == list:
            for k in key:
                if not k in self._apikey:
                    self._apikey.append(k)

    def delkey(self, key):
        """
        Removes a key (unregister ?)
        """
        if type(key) == str:
            if key in self._apikey:
                self._apikey.remove(key)
        elif type(key) == list:
            for k in key:
                if key in self._apikey:
                    self._apikey.remove(k)

    def developerkey(self, developerkey):
        "Sets the developer key (and check it has the good length)"
        if type(developerkey) == str and len(developerkey) == 48:
            self._developerkey = developerkey

    def push(self, application="", event="", description="", url="", contenttype=None, priority=0, batch_mode=False,
             html=False):
        """Pushes a message on the registered API keys.
        """
        datas = {
            'application': application[:256].encode('utf8'),
            'event': event[:1024].encode('utf8'),
            'description': description[:10000].encode('utf8'),
            'priority': priority
        }

        if url:
            datas['url'] = url[:512]

        if contenttype == "text/html" or html == True:  # Currently only accepted content type
            datas['content-type'] = "text/html"

        if self._developerkey:
            datas['developerkey'] = self._developerkey

        results = {}

        if not batch_mode:
            for key in self._apikey:
                datas['apikey'] = key
                res = self.callapi('POST', datas)
                results[key] = res
        else:
            datas['apikey'] = ",".join(self._apikey)
            res = self.callapi('POST', datas)
            results[datas['apikey']] = res

        return results

    def callapi(self, method, args):
        headers = {'User-Agent': sickrage.app.user_agent}
        if method == "POST":
            headers['Content-type'] = "application/x-www-form-urlencoded"

        resp = request(method, url=urljoin(self.url, '/publicapi/notify'), headers=headers, data=urlencode(args))

        try:
            res = self._parse_reponse(resp.content)
        except Exception as e:
            res = {'type': "pynmaerror",
                   'code': 600,
                   'message': str(e)
                   }
            pass

        return res

    def _parse_reponse(self, response):
        root = parseString(response).firstChild
        for elem in root.childNodes:
            if elem.nodeType == elem.TEXT_NODE: continue
            if elem.tagName == 'success':
                res = dict(list(elem.attributes.items()))
                res['message'] = ""
                res['type'] = elem.tagName
                return res
            if elem.tagName == 'error':
                res = dict(list(elem.attributes.items()))
                res['message'] = elem.firstChild.nodeValue
                res['type'] = elem.tagName
                return res

    def test_notify(self, nma_api, nma_priority):
        return self._sendNMA(nma_api, nma_priority, event="Test", message="Testing NMA settings from SiCKRAGE",
                             force=True)

    def notify_snatch(self, ep_name):
        if sickrage.app.config.nma_notify_onsnatch:
            self._sendNMA(event=self.notifyStrings[self.NOTIFY_SNATCH], message=ep_name)

    def notify_download(self, ep_name):
        if sickrage.app.config.nma_notify_ondownload:
            self._sendNMA(event=self.notifyStrings[self.NOTIFY_DOWNLOAD], message=ep_name)

    def notify_subtitle_download(self, ep_name, lang):
        if sickrage.app.config.nma_notify_onsubtitledownload:
            self._sendNMA(event=self.notifyStrings[self.NOTIFY_SUBTITLE_DOWNLOAD], message=ep_name + ": " + lang)

    def notify_version_update(self, new_version="??"):
        if sickrage.app.config.use_nma:
            update_text = self.notifyStrings[self.NOTIFY_GIT_UPDATE_TEXT]
            title = self.notifyStrings[self.NOTIFY_GIT_UPDATE]
            self._sendNMA(event=title, message=update_text + new_version)

    def _sendNMA(self, nma_api=None, nma_priority=None, event=None, message=None, force=False):

        title = 'SiCKRAGE'

        if not sickrage.app.config.use_nma and not force:
            return False

        if nma_api is None:
            nma_api = sickrage.app.config.nma_api

        if nma_priority is None:
            nma_priority = sickrage.app.config.nma_priority

        batch = False

        keys = nma_api.split(',')
        self.addkey(keys)

        if len(keys) > 1: batch = True

        sickrage.app.log.debug(
            "NMA: Sending notice with details: event=\"%s\", message=\"%s\", priority=%s, batch=%s" % (
                event, message, nma_priority, batch))

        response = self.push(
            application=title,
            event=event,
            description=message,
            priority=nma_priority,
            batch_mode=batch
        )

        if not response[nma_api]['code'] == '200':
            sickrage.app.log.warning('Could not send notification to NotifyMyAndroid')
            return False
        else:
            sickrage.app.log.info("NMA: Notification sent to NotifyMyAndroid")
            return True



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################

import datetime

from sqlalchemy import *


def upgrade(migrate_engine):
    meta = MetaData(bind=migrate_engine)
    history = Table('history', meta, autoload=True)

    history.c.date.alter(type=String(32))

    date_format = '%Y%m%d%H%M%S'
    with migrate_engine.begin() as conn:
        for row in migrate_engine.execute(history.select()):
            conn.execute(history.update().where(history.c.id == row.id).values(date=datetime.datetime.strptime(str(row.date), date_format)))

    history.c.date.alter(type=DateTime)


def downgrade(migrate_engine):
    meta = MetaData(bind=migrate_engine)
    history = Table('history', meta, autoload=True)

    history.c.date.alter(type=String(32))

    with migrate_engine.begin() as conn:
        for row in migrate_engine.execute(history.select()):
            conn.execute(history.update().where(history.c.id == row.id).values(airdate=str(row.date.toordinal())))

    history.c.date.alter(type=Integer)



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca/
# Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.


import datetime
import sys
from collections import deque

from sickrage.core.common import Quality, dateTimeFormat


class SearchResult(object):
    """
    Represents a search result from an indexer.
    """

    def __init__(self, season, episodes):
        self.provider = None

        # release name
        self.name = ""

        # release show ID
        self.show_id = None

        # URL to the NZB/torrent file
        self.url = ""

        # used by some providers to store extra info associated with the result
        self.extraInfo = []

        # season that this result is associated with
        self.season = season

        # list of episodes that this result is associated with
        self.episodes = episodes

        # quality of the release
        self.quality = Quality.UNKNOWN

        # size of the release (-1 = n/a)
        self.size = -1

        # seeders of the release
        self.seeders = -1

        # leechers of the release
        self.leechers = -1

        # update date
        self.date = None

        # release group
        self.release_group = ""

        # version
        self.version = -1

        # hash
        self.hash = None

        # content
        self.content = None

        # ratio
        self.ratio = None

        # result type
        self.type = ''

        # dict of files and their sizes
        self.files = {}

    def __str__(self):

        if self.provider is None:
            return "Invalid provider, unable to print self"

        myString = self.provider.name + " @ " + self.url + "\n"

        myString += "Extra Info:\n"
        myString += "  ".join(self.extraInfo) + "\n"

        myString += "Season: " + str(self.season) + "\n"

        myString += "Episodes:\n"
        myString += "  ".join(map(str, self.episodes)) + "\n"

        myString += "Quality: " + Quality.qualityStrings[self.quality] + "\n"
        myString += "Name: " + self.name + "\n"
        myString += "Size: " + str(self.size) + "\n"
        myString += "Release Group: " + str(self.release_group) + "\n"

        return myString


class NZBSearchResult(SearchResult):
    """
    Regular NZB result with an URL to the NZB
    """

    def __init__(self, season, episodes):
        super(NZBSearchResult, self).__init__(season, episodes)
        self.type = "nzb"


class NZBDataSearchResult(SearchResult):
    """
    NZB result where the actual NZB XML data is stored in the extraInfo
    """

    def __init__(self, season, episodes):
        super(NZBDataSearchResult, self).__init__(season, episodes)
        self.type = "nzbdata"


class TorrentSearchResult(SearchResult):
    """
    Torrent result with an URL to the torrent
    """

    def __init__(self, season, episodes):
        super(TorrentSearchResult, self).__init__(season, episodes)
        self.type = "torrent"


class UIError(object):
    """
    Represents an error to be displayed in the web UI.
    """

    def __init__(self, message):
        self.time = datetime.datetime.now().strftime(dateTimeFormat)
        self.title = sys.exc_info()[-2] or message
        self.message = message


class UIWarning(object):
    """
    Represents an error to be displayed in the web UI.
    """

    def __init__(self, message):
        self.time = datetime.datetime.now().strftime(dateTimeFormat)
        self.title = sys.exc_info()[-2] or message
        self.message = message


class ErrorViewer(object):
    """
    Keeps a static list of UIErrors to be displayed on the UI and allows
    the list to be cleared.
    """

    def __init__(self):
        self.errors = deque(maxlen=100)

    def add(self, error, ui=False):
        self.errors += [(error, UIError(error))[ui]]

    def clear(self):
        self.errors.clear()

    def get(self):
        return self.errors

    def count(self):
        return len(self.errors)


class WarningViewer(object):
    """
    Keeps a static list of (warning) UIErrors to be displayed on the UI and allows
    the list to be cleared.
    """

    def __init__(self):
        self.warnings = deque(maxlen=100)

    def add(self, warning, ui=False):
        self.warnings += [(warning, UIWarning(warning))[ui]]

    def clear(self):
        self.warnings.clear()

    def get(self):
        return self.warnings

    def count(self):
        return len(self.warnings)



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################


import base64
import json
import socket
import time
import uuid
from urllib import parse
from urllib.parse import unquote_plus, urlencode, unquote
from xml.etree import ElementTree

import sickrage
from sickrage.core.websession import WebSession
from sickrage.notifiers import Notifiers


class KODINotifier(Notifiers):
    sr_logo_url = 'https://www.sickrage.ca/favicon.ico'

    def __init__(self):
        super(KODINotifier, self).__init__()
        self.name = 'kodi'

    def _get_kodi_version(self, host, username, password):
        """Returns KODI JSON-RPC API version (odd # = dev, even # = stable)

        Sends a request to the KODI host using the JSON-RPC to determine if
        the legacy API or if the JSON-RPC API functions should be used.

        Fallback to testing legacy HTTPAPI before assuming it is just a badly configured host.

        Args:
            host: KODI webserver host:port
            username: KODI webserver username
            password: KODI webserver password

        Returns:
            Returns API number or False

            List of possible known values:
                API | KODI Version
               -----+---------------
                 2  | v10 (Dharma)
                 3  | (pre Eden)
                 4  | v11 (Eden)
                 5  | (pre Frodo)
                 6  | v12 (Frodo) / v13 (Gotham)

        """

        # since we need to maintain python 2.5 compatability we can not pass a timeout delay to urllib2 directly (python 2.6+)
        # override socket timeout to reduce delay for this call alone
        socket.setdefaulttimeout(10)

        checkCommand = {"jsonrpc": "2.0", "method": "JSONRPC.Version", "id": uuid.uuid4().hex}
        result = self._send_to_kodi_json(checkCommand, host, username, password)

        # revert back to default socket timeout
        socket.setdefaulttimeout(sickrage.app.config.socket_timeout)

        if not result or result is False or 'error' in result:
            # fallback to legacy HTTPAPI method
            testCommand = {'command': 'Help'}
            request = self._send_to_kodi(testCommand, host, username, password)
            if request:
                # return a fake version number, so it uses the legacy method
                return 1
            else:
                return False

        return result["result"]["version"]["major"]

    def _notify_kodi(self, message, title="SiCKRAGE", host=None, username=None, password=None, force=False):
        """Internal wrapper for the notify_snatch and notify_download functions

        Detects JSON-RPC version then branches the logic for either the JSON-RPC or legacy HTTP API methods.

        Args:
            message: Message body of the notice to send
            title: Title of the notice to send
            host: KODI webserver host:port
            username: KODI webserver username
            password: KODI webserver password
            force: Used for the Test method to override config saftey checks

        Returns:
            Returns a list results in the format of host:ip:result
            The result will either be 'OK' or False, this is used to be parsed by the calling function.

        """

        # fill in omitted parameters
        if not host:
            host = sickrage.app.config.kodi_host
        if not username:
            username = sickrage.app.config.kodi_username
        if not password:
            password = sickrage.app.config.kodi_password

        # suppress notifications if the notifier is disabled but the notify options are checked
        if not sickrage.app.config.use_kodi and not force:
            sickrage.app.log.debug("Notification for KODI not enabled, skipping this notification")
            return False

        result = ''
        for curHost in [x.strip() for x in host.split(",")]:
            sickrage.app.log.debug("Sending KODI notification to '" + curHost + "' - " + message)

            kodiapi = self._get_kodi_version(curHost, username, password)
            if kodiapi:
                if kodiapi <= 4:
                    sickrage.app.log.debug("Detected KODI version <= 11, using KODI HTTP API")
                    command = {'command': 'ExecBuiltIn',
                               'parameter': 'Notification({},{})'.format(title.encode("utf-8"), message.encode("utf-8"))}
                    notifyResult = self._send_to_kodi(command, curHost, username, password)
                    if notifyResult:
                        result += curHost + ':' + str(notifyResult)
                else:
                    sickrage.app.log.debug("Detected KODI version >= 12, using KODI JSON API")

                    command = {"jsonrpc": "2.0",
                               "method": "GUI.ShowNotification",
                               "params": {"title": title, "message": message, "image": self.sr_logo_url},
                               "id": uuid.uuid4().hex}

                    notifyResult = self._send_to_kodi_json(command, curHost, username, password)
                    if notifyResult and notifyResult.get('result'):
                        result += curHost + ':' + notifyResult["result"]
            else:
                if sickrage.app.config.kodi_always_on or force:
                    sickrage.app.log.warning("Failed to detect KODI version for '" + curHost + "', check configuration and try again.")
                result += curHost + ':False'

        return result

    def _send_update_library(self, host, showName=None):
        """Internal wrapper for the update library function to branch the logic for JSON-RPC or legacy HTTP API

        Checks the KODI API version to branch the logic to call either the legacy HTTP API or the newer JSON-RPC over HTTP methods.

        Args:
            host: KODI webserver host:port
            showName: Name of a TV show to specifically target the library update for

        Returns:
            Returns True or False, if the update was successful

        """

        sickrage.app.log.debug("Sending request to update library for KODI host: '" + host + "'")

        kodiapi = self._get_kodi_version(host, sickrage.app.config.kodi_username,
                                         sickrage.app.config.kodi_password)
        if kodiapi:
            if kodiapi <= 4:
                # try to update for just the show, if it fails, do full update if enabled
                if not self._update_library(host, showName) and sickrage.app.config.kodi_update_full:
                    sickrage.app.log.debug("Single show update failed, falling back to full update")
                    return self._update_library(host)
                else:
                    return True
            else:
                # try to update for just the show, if it fails, do full update if enabled
                if not self._update_library_json(host, showName) and sickrage.app.config.kodi_update_full:
                    sickrage.app.log.debug("Single show update failed, falling back to full update")
                    return self._update_library_json(host)
                else:
                    return True
        elif sickrage.app.config.kodi_always_on:
            sickrage.app.log.warning(
                "Failed to detect KODI version for '" + host + "', check configuration and try again.")

        return False

    # #############################################################################
    # Legacy HTTP API (pre KODI 12) methods
    ##############################################################################

    def _send_to_kodi(self, command, host=None, username=None, password=None):
        """Handles communication to KODI servers via HTTP API

        Args:
            command: Dictionary of field/data pairs, encoded via urllib and passed to the KODI API via HTTP
            host: KODI webserver host:port
            username: KODI webserver username
            password: KODI webserver password

        Returns:
            Returns response.result for successful commands or False if there was an error

        """

        # fill in omitted parameters
        if not username:
            username = sickrage.app.config.kodi_username
        if not password:
            password = sickrage.app.config.kodi_password

        if not host:
            sickrage.app.log.warning('No KODI host passed, aborting update')
            return False

        enc_command = urlencode(command)
        sickrage.app.log.debug("KODI encoded API command: " + enc_command)

        url = 'http://%s/kodiCmds/kodiHttp/?%s' % (host, enc_command)

        headers = {}

        # if we have a password, use authentication
        if password:
            authheader = "Basic {}".format(base64.b64encode(bytes('{}:{}'.format(username, password).replace('\n', ''), 'utf-8')).decode('ascii'))
            headers["Authorization"] = authheader
            sickrage.app.log.debug("Contacting KODI (with auth header) via url: " + url)
        else:
            sickrage.app.log.debug("Contacting KODI via url: " + url)

        try:
            result = WebSession().get(url, headers=headers).text
        except Exception as e:
            sickrage.app.log.debug("Couldn't contact KODI HTTP at %r : %r" % (url, e))
            return False

        sickrage.app.log.debug("KODI HTTP response: " + result.replace('\n', ''))
        return result

    def _update_library(self, host=None, showName=None):
        """Handles updating KODI host via HTTP API

        Attempts to update the KODI video library for a specific tv show if passed,
        otherwise update the whole library if enabled.

        Args:
            host: KODI webserver host:port
            showName: Name of a TV show to specifically target the library update for

        Returns:
            Returns True or False

        """

        if not host:
            sickrage.app.log.warning('No KODI host passed, aborting update')
            return False

        sickrage.app.log.debug("Updating KODI library via HTTP method for host: " + host)

        # if we're doing per-show
        if showName:
            sickrage.app.log.debug("Updating library in KODI via HTTP method for show " + showName)

            pathSql = 'select path.strPath from path, tvshow, tvshowlinkpath where ' \
                      'tvshow.c00 = "{0:s}" and tvshowlinkpath.idShow = tvshow.idShow ' \
                      'and tvshowlinkpath.idPath = path.idPath'.format(showName)

            # use this to get xml back for the path lookups
            xmlCommand = {
                'command': 'SetResponseFormat(webheader;false;webfooter;false;header;<xml>;footer;</xml>;opentag;<tag>;closetag;</tag>;closefinaltag;false)'}
            # sql used to grab path(s)
            sqlCommand = {'command': 'QueryVideoDatabase(%s)' % (pathSql)}
            # set output back to default
            resetCommand = {'command': 'SetResponseFormat()'}

            # set xml response format, if this fails then don't bother with the rest
            request = self._send_to_kodi(xmlCommand, host)
            if not request:
                return False

            sqlXML = self._send_to_kodi(sqlCommand, host)
            request = self._send_to_kodi(resetCommand, host)

            if not sqlXML:
                sickrage.app.log.debug("Invalid response for " + showName + " on " + host)
                return False

            encSqlXML = parse.quote(sqlXML, ':\\/<>')
            try:
                et = ElementTree.fromstring(encSqlXML)
            except SyntaxError as e:
                sickrage.app.log.error("Unable to parse XML returned from KODI: {}".format(e))
                return False

            paths = et.findall('.//field')

            if not paths:
                sickrage.app.log.debug("No valid paths found for " + showName + " on " + host)
                return False

            for path in paths:
                # we do not need it double-encoded, gawd this is dumb
                unEncPath = unquote(path.text)
                sickrage.app.log.debug("KODI Updating " + showName + " on " + host + " at " + unEncPath)
                updateCommand = {'command': 'ExecBuiltIn', 'parameter': 'KODI.updatelibrary(video, %s)' % (unEncPath)}
                request = self._send_to_kodi(updateCommand, host)
                if not request:
                    sickrage.app.log.warning(
                        "Update of show directory failed on " + showName + " on " + host + " at " + unEncPath)
                    return False
                # sleep for a few seconds just to be sure kodi has a chance to finish each directory
                if len(paths) > 1:
                    time.sleep(5)
        # do a full update if requested
        else:
            sickrage.app.log.debug("Doing Full Library KODI update on host: " + host)
            updateCommand = {'command': 'ExecBuiltIn', 'parameter': 'KODI.updatelibrary(video)'}
            request = self._send_to_kodi(updateCommand, host)

            if not request:
                sickrage.app.log.warning("KODI Full Library update failed on: " + host)
                return False

        return True

    ##############################################################################
    # JSON-RPC API (KODI 12+) methods
    ##############################################################################

    def _send_to_kodi_json(self, command, host=None, username=None, password=None):
        """Handles communication to KODI servers via JSONRPC

        Args:
            command: Dictionary of field/data pairs, encoded via urllib and passed to the KODI JSON-RPC via HTTP
            host: KODI webserver host:port
            username: KODI webserver username
            password: KODI webserver password

        Returns:
            Returns response.result for successful commands or False if there was an error

        """

        # fill in omitted parameters
        if not username:
            username = sickrage.app.config.kodi_username
        if not password:
            password = sickrage.app.config.kodi_password

        if not host:
            sickrage.app.log.warning('No KODI host passed, aborting update')
            return False

        sickrage.app.log.debug("KODI JSON command: {!r}".format(command))

        url = 'http://%s/jsonrpc' % host

        headers = {"Content-type": "application/json"}

        # if we have a password, use authentication
        if password:
            authheader = "Basic {}".format(base64.b64encode(bytes('{}:{}'.format(username, password).replace('\n', ''), 'utf-8')).decode('ascii'))
            headers["Authorization"] = authheader
            sickrage.app.log.debug("Contacting KODI (with auth header) via url: " + url)
        else:
            sickrage.app.log.debug("Contacting KODI via url: " + url)

        try:
            result = WebSession().post(url, json=command, headers=headers).json()
            sickrage.app.log.debug("KODI JSON response: " + str(result))
            return result
        except Exception as e:
            if sickrage.app.config.kodi_always_on:
                sickrage.app.log.warning("Warning: Couldn't contact KODI JSON API at " + url + " {}".format(e))

        return False

    def _update_library_json(self, host=None, showName=None):
        """Handles updating KODI host via HTTP JSON-RPC

        Attempts to update the KODI video library for a specific tv show if passed,
        otherwise update the whole library if enabled.

        Args:
            host: KODI webserver host:port
            showName: Name of a TV show to specifically target the library update for

        Returns:
            Returns True or False

        """

        if not host:
            sickrage.app.log.warning('No KODI host passed, aborting update')
            return False

        sickrage.app.log.debug("Updating KODI library via JSON method for host: " + host)

        # if we're doing per-show
        if showName:
            showName = unquote_plus(showName)
            tvshowid = -1
            path = ''

            sickrage.app.log.debug("Updating library in KODI via JSON method for show " + showName)

            # let's try letting kodi filter the shows
            showsCommand = {"jsonrpc": "2.0", "method": "VideoLibrary.GetTVShows",
                            "params": {"filter": {"field": "title", "operator": "is", "value": showName}, "properties": ["title"]},
                            "id": uuid.uuid4().hex}

            # get tvshowid by showName
            showsResponse = self._send_to_kodi_json(showsCommand, host)

            if showsResponse and "result" in showsResponse and "tvshows" in showsResponse["result"]:
                shows = showsResponse["result"]["tvshows"]
            else:
                # fall back to retrieving the entire show list
                showsCommand = {"jsonrpc": "2.0", "method": "VideoLibrary.GetTVShows", "id": uuid.uuid4().hex}
                showsResponse = self._send_to_kodi_json(showsCommand, host)

                if showsResponse and "result" in showsResponse and "tvshows" in showsResponse["result"]:
                    shows = showsResponse["result"]["tvshows"]
                else:
                    sickrage.app.log.debug("KODI: No tvshows in KODI TV show list")
                    return False

            for show in shows:
                if ("label" in show and show["label"] == showName) or ("title" in show and show["title"] == showName):
                    tvshowid = show["tvshowid"]
                    # set the path is we have it already
                    if "file" in show:
                        path = show["file"]

                    break

            # this can be big, so free some memory
            del shows

            # we didn't find the show (exact match), thus revert to just doing a full update if enabled
            if tvshowid == -1:
                sickrage.app.log.debug('Exact show name not matched in KODI TV show list')
                return False

            # lookup tv-show path if we don't already know it
            if not len(path):
                pathCommand = {"jsonrpc": "2.0",
                               "method": "VideoLibrary.GetTVShowDetails",
                               "params": {"tvshowid": tvshowid, "properties": ["file"]},
                               "id": uuid.uuid4().hex}

                pathResponse = self._send_to_kodi_json(pathCommand, host)

                path = pathResponse["result"]["tvshowdetails"]["file"]

            sickrage.app.log.debug(
                "Received Show: " + showName + " with ID: " + str(tvshowid) + " Path: " + path)

            if not len(path):
                sickrage.app.log.warning("No valid path found for " + showName + " with ID: " + str(tvshowid) + " on " + host)
                return False

            sickrage.app.log.debug("KODI Updating " + showName + " on " + host + " at " + path)
            updateCommand = {"jsonrpc": "2.0", "method": "VideoLibrary.Scan", "params": {"directory": path}, "id": uuid.uuid4().hex}
            request = self._send_to_kodi_json(updateCommand, host)
            if request is False:
                sickrage.app.log.warning("Update of show directory failed on " + showName + " on " + host + " at " + path)
                return False

            # catch if there was an error in the returned request
            for r in request:
                if 'error' in r:
                    sickrage.app.log.warning("Error while attempting to update show directory for " + showName + " on " + host + " at " + path)
                    return False

        # do a full update if requested
        else:
            sickrage.app.log.debug("Doing Full Library KODI update on host: " + host)
            updateCommand = {"jsonrpc": "2.0", "method": "VideoLibrary.Scan", "id": uuid.uuid4().hex}
            request = self._send_to_kodi_json(updateCommand, host)

            if not request:
                sickrage.app.log.warning("KODI Full Library update failed on: " + host)
                return False

        return True

    ##############################################################################
    # Public functions which will call the JSON or Legacy HTTP API methods
    ##############################################################################

    def notify_snatch(self, ep_name):
        if sickrage.app.config.kodi_notify_onsnatch:
            self._notify_kodi(ep_name, self.notifyStrings[self.NOTIFY_SNATCH])

    def notify_download(self, ep_name):
        if sickrage.app.config.kodi_notify_ondownload:
            self._notify_kodi(ep_name, self.notifyStrings[self.NOTIFY_DOWNLOAD])

    def notify_subtitle_download(self, ep_name, lang):
        if sickrage.app.config.kodi_notify_onsubtitledownload:
            self._notify_kodi(ep_name + ": " + lang, self.notifyStrings[self.NOTIFY_SUBTITLE_DOWNLOAD])

    def notify_version_update(self, new_version="??"):
        if sickrage.app.config.use_kodi:
            update_text = self.notifyStrings[self.NOTIFY_GIT_UPDATE_TEXT]
            title = self.notifyStrings[self.NOTIFY_GIT_UPDATE]
            self._notify_kodi(update_text + new_version, title)

    def test_notify(self, host, username, password):
        return self._notify_kodi("Testing KODI notifications from SiCKRAGE", "Test Notification", host, username, password, force=True)

    def update_library(self, showName=None):
        """Public wrapper for the update library functions to branch the logic for JSON-RPC or legacy HTTP API

        Checks the KODI API version to branch the logic to call either the legacy HTTP API or the newer JSON-RPC over HTTP methods.
        Do the ability of accepting a list of hosts deliminated by comma, only one host is updated, the first to respond with success.
        This is a workaround for SQL backend users as updating multiple clients causes duplicate entries.
        Future plan is to revist how we store the host/ip/username/pw/options so that it may be more flexible.

        Args:
            showName: Name of a TV show to specifically target the library update for

        Returns:
            Returns True or False

        """

        if sickrage.app.config.use_kodi and sickrage.app.config.kodi_update_library:
            if not sickrage.app.config.kodi_host:
                sickrage.app.log.debug("No KODI hosts specified, check your settings")
                return False

            # either update each host, or only attempt to update until one successful result
            result = 0
            for host in [x.strip() for x in sickrage.app.config.kodi_host.split(",")]:
                if self._send_update_library(host, showName):
                    if sickrage.app.config.kodi_update_onlyfirst:
                        sickrage.app.log.debug("Successfully updated '" + host + "', stopped sending update library commands.")
                        return True
                else:
                    if sickrage.app.config.kodi_always_on:
                        sickrage.app.log.warning("Failed to detect KODI version for '" + host + "', check configuration and try again.")
                    result += 1

            # needed for the 'update kodi' submenu command
            # as it only cares of the final result vs the individual ones
            if result == 0:
                return True
            else:
                return False



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################


import os
import shutil
import stat

import rarfile
from sqlalchemy import or_

import sickrage
from sickrage.core.common import Quality
from sickrage.core.databases.main import MainDB
from sickrage.core.exceptions import EpisodePostProcessingFailedException, \
    FailedPostProcessingFailedException, NoFreeSpaceException
from sickrage.core.helpers import is_media_file, is_rar_file, is_hidden_folder, real_path, is_torrent_or_nzb_file, \
    is_sync_file, get_extension
from sickrage.core.nameparser import InvalidNameException, InvalidShowException, \
    NameParser
from sickrage.core.processors import failed_processor, post_processor
from sickrage.core.tv.show.helpers import get_show_list


class ProcessResult(object):
    def __init__(self, path, process_method=None, process_type='auto'):
        self._output = []
        self._path = path
        self.process_method = process_method or sickrage.app.config.process_method
        self.process_type = process_type
        self.video_files = []
        self.missed_files = []
        self.result = True
        self.succeeded = True

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, value):
        dir_name = None
        if os.path.isdir(value):
            dir_name = os.path.realpath(value)
            self.log("Processing in folder {0}".format(dir_name), sickrage.app.log.DEBUG)
        elif all([sickrage.app.config.tv_download_dir,
                  os.path.isdir(sickrage.app.config.tv_download_dir),
                  os.path.normpath(value) == os.path.normpath(sickrage.app.config.tv_download_dir)]):
            dir_name = os.path.join(sickrage.app.config.tv_download_dir,
                                    os.path.abspath(value).split(os.path.sep)[-1])
            self.log("Trying to use folder: {0} ".format(dir_name), sickrage.app.log.DEBUG)
        else:
            self.log("Unable to figure out what folder to process. "
                     "If your downloader and SiCKRAGE aren't on the same PC "
                     "make sure you fill out your TV download dir in the config.",
                     sickrage.app.log.DEBUG)

        self._path = dir_name

    @property
    def output(self):
        return '\n'.join(self._output)

    def log(self, message, level=None):
        sickrage.app.log.log(level or sickrage.app.log.INFO, message)
        self._output.append(message)

    def clear_log(self):
        self._output = []

    def delete_folder(self, folder, check_empty=True):
        """
        Removes a folder from the filesystem

        :param folder: Path to folder to remove
        :param check_empty: Boolean, check if the folder is empty before removing it, defaults to True
        :return: True on success, False on failure
        """

        # check if it's a folder
        if not os.path.isdir(folder):
            return False

        # check if it isn't TV_DOWNLOAD_DIR
        if sickrage.app.config.tv_download_dir:
            if real_path(folder) == real_path(sickrage.app.config.tv_download_dir):
                return False

        # check if it's empty folder when wanted checked
        try:
            if check_empty:
                check_files = os.listdir(folder)
                if check_files:
                    sickrage.app.log.info(
                        "Not deleting folder {} found the following files: {}".format(folder, check_files))
                    return False

                sickrage.app.log.info("Deleting folder (if it's empty): " + folder)
                shutil.rmtree(folder)
            else:
                sickrage.app.log.info("Deleting folder: " + folder)
                shutil.rmtree(folder)
        except (OSError, IOError) as e:
            sickrage.app.log.warning("Warning: unable to delete folder: {}: {}".format(folder, e))
            return False

        return True

    def delete_files(self, processPath, notwantedFiles, force=False):
        """
        Remove files from filesystem

        :param processPath: path to process
        :param notwantedFiles: files we do not want
        :param force: Boolean, force deletion, defaults to false
        """

        if not self.result and force:
            self.log("Forcing deletion of files, even though last result was not success",
                     sickrage.app.log.DEBUG)
        elif not self.result:
            return

        # Delete all file not needed
        for cur_file in notwantedFiles:
            cur_file_path = os.path.join(processPath, cur_file)

            if not os.path.isfile(cur_file_path):
                continue  # Prevent error when a notwantedfiles is an associated files

            self.log("Deleting file %s" % cur_file, sickrage.app.log.DEBUG)

            # check first the read-only attribute
            file_attribute = os.stat(cur_file_path)[0]
            if not file_attribute & stat.S_IWRITE:
                # File is read-only, so make it writeable
                self.log("Changing ReadOnly Flag for file {}".format(cur_file),
                         sickrage.app.log.DEBUG)
                try:
                    os.chmod(cur_file_path, stat.S_IWRITE)
                except OSError as e:
                    self.log("Cannot change permissions of {}: {}".format(cur_file, e.strerror),
                             sickrage.app.log.DEBUG)
            try:
                os.remove(cur_file_path)
            except OSError as e:
                self.log("Unable to delete file {}: {}".format(cur_file, e.strerror),
                         sickrage.app.log.DEBUG)

    def process(self, nzbName=None, force=False, is_priority=None, delete_on=False, failed=False):
        """
        Scans through the files in dir_name and processes whatever media files it finds

        :param nzbName: The NZB name which resulted in this folder being downloaded
        :param force: True to postprocess already postprocessed files
        :param is_priority: whether to replace the file even if it exists at higher quality
        :param delete_on: delete files and folders after they are processed (always happens with move and auto combination)
        :param failed: Boolean for whether or not the download failed
        """

        self.clear_log()

        directories_from_rars = set()

        # If we have a release name (probably from nzbToMedia), and it is a rar/video, only process that file
        if nzbName and (is_media_file(nzbName) or is_rar_file(nzbName)):
            self.log("Processing {}".format(nzbName), sickrage.app.log.INFO)
            generator_to_use = [(self.path, [], [nzbName])]
        else:
            self.log("Processing {}".format(self.path), sickrage.app.log.INFO)
            generator_to_use = os.walk(self.path, followlinks=sickrage.app.config.processor_follow_symlinks)

        rar_files = []
        for current_directory, directory_names, file_names in generator_to_use:
            self.result = True

            file_names = [f for f in file_names if not is_torrent_or_nzb_file(f)]
            rar_files = [x for x in file_names if is_rar_file(os.path.join(current_directory, x))]
            if rar_files:
                extracted_directories = self.unrar(current_directory, rar_files, force)
                if extracted_directories:
                    for extracted_directory in extracted_directories:
                        if extracted_directory.split(current_directory)[-1] not in directory_names:
                            self.log(
                                "Adding extracted directory to the list of directories to process: {0}".format(
                                    extracted_directory), sickrage.app.log.DEBUG
                            )
                            directories_from_rars.add(extracted_directory)

            if not self.validateDir(current_directory, nzbName, failed):
                continue

            video_files = list(filter(is_media_file, file_names))
            if video_files:
                try:
                    self.process_media(current_directory, video_files, nzbName, self.process_method, force, is_priority)
                except NoFreeSpaceException:
                    continue
            else:
                self.result = False

            # Delete all file not needed and avoid deleting files if Manual PostProcessing
            if not (self.process_method == "move" and self.result) or (self.process_type == "manual" and not delete_on):
                continue

            # Check for unwanted files
            unwanted_files = list(
                filter(
                    lambda x: x not in video_files and get_extension(x) not in sickrage.app.config.allowed_extensions,
                    file_names)
            )

            if unwanted_files:
                self.log("Found unwanted files: {0}".format(unwanted_files), sickrage.app.log.DEBUG)

            self.delete_folder(os.path.join(current_directory, '@eaDir'), False)
            self.delete_files(current_directory, unwanted_files)
            if self.delete_folder(current_directory, check_empty=not delete_on):
                self.log("Deleted folder: {0}".format(current_directory), sickrage.app.log.DEBUG)

        method_fallback = ('move', self.process_method)[self.process_method in ('move', 'copy')]

        delete_rar_contents = any([sickrage.app.config.delrarcontents and self.process_type != 'manual',
                                   not sickrage.app.config.delrarcontents and self.process_type == 'auto' and method_fallback == 'move',
                                   self.process_type == 'manual' and delete_on])

        for directory_from_rar in directories_from_rars:
            ProcessResult(directories_from_rars, self.process_method, self.process_type).process(
                nzbName=os.path.basename(directory_from_rar),
                force=force,
                is_priority=is_priority,
                delete_on=delete_rar_contents,
                failed=failed
            )

            # Delete rar file only if the extracted dir was successfully processed
            if self.process_type == 'auto' and method_fallback == 'move' or self.process_type == 'manual' and delete_on:
                this_rar = [rar_file for rar_file in rar_files if
                            os.path.basename(directory_from_rar) == rar_file.rpartition('.')[0]]
                self.delete_files(self.path, this_rar)

        self.log(("Processing Failed", "Successfully processed")[self.succeeded],
                 (sickrage.app.log.WARNING, sickrage.app.log.INFO)[self.succeeded])

        if self.missed_files:
            self.log("Some items were not processed.")
            for missed_file in self.missed_files:
                self.log(missed_file)

        return self.output

    def validateDir(self, process_path, release_name, failed):
        """
        Check if directory is valid for processing

        :param process_path: Directory to check
        :param release_name: Original NZB/Torrent name
        :param failed: Previously failed objects
        :return: True if dir is valid for processing, False if not
        """

        self.log("Processing folder " + process_path, sickrage.app.log.DEBUG)

        upper_name = os.path.basename(process_path).upper()
        if upper_name.startswith('_FAILED_') or upper_name.endswith('_FAILED_'):
            self.log("The directory name indicates it failed to extract.", sickrage.app.log.DEBUG)
            failed = True
        elif upper_name.startswith('_UNDERSIZED_') or upper_name.endswith('_UNDERSIZED_'):
            self.log(
                "The directory name indicates that it was previously rejected for being undersized.",
                sickrage.app.log.DEBUG)
            failed = True
        elif upper_name.startswith('_UNPACK') or upper_name.endswith('_UNPACK'):
            self.log(
                "The directory name indicates that this release is in the process of being unpacked.",
                sickrage.app.log.DEBUG)
            self.missed_files.append("{0} : Being unpacked".format(process_path))
            return False

        if failed:
            self.process_failed(process_path, release_name)
            self.missed_files.append("{0} : Failed download".format(process_path))
            return False

        if sickrage.app.config.tv_download_dir and real_path(process_path) != real_path(
                sickrage.app.config.tv_download_dir) and is_hidden_folder(process_path):
            self.log("Ignoring hidden folder: {0}".format(process_path), sickrage.app.log.DEBUG)
            self.missed_files.append("{0} : Hidden folder".format(process_path))
            return False

        # make sure the dir isn't inside a show dir
        for show in get_show_list():
            if process_path.lower().startswith(os.path.realpath(show.location).lower() + os.sep) or \
                    process_path.lower() == os.path.realpath(show.location).lower():
                self.log("Cannot process an episode that's already been moved to its show dir, skipping " + process_path, sickrage.app.log.WARNING)
                return False

        for current_directory, directory_names, file_names in os.walk(process_path, topdown=False,
                                                                      followlinks=sickrage.app.config.processor_follow_symlinks):
            sync_files = list(filter(is_sync_file, file_names))
            if sync_files and sickrage.app.config.postpone_if_sync_files:
                self.log("Found temporary sync files: {0} in path: {1}".format(sync_files,
                                                                               os.path.join(
                                                                                   process_path,
                                                                                   sync_files[
                                                                                       0])))
                self.log("Skipping post processing for folder: {0}".format(process_path))
                self.missed_files.append("{0} : Sync files found".format(os.path.join(process_path, sync_files[0])))
                continue

            found_files = list(filter(is_media_file, file_names))
            if sickrage.app.config.unpack == 1:
                found_files += list(filter(is_rar_file, file_names))

            if current_directory != sickrage.app.config.tv_download_dir and found_files:
                found_files.append(os.path.basename(current_directory))

            for found_file in found_files:
                try:
                    NameParser().parse(found_file, cache_result=False)
                except (InvalidNameException, InvalidShowException) as e:
                    pass
                else:
                    return True

        self.log("Folder {} : No processable items found in folder".format(process_path),
                 sickrage.app.log.DEBUG)
        return False

    def unrar(self, path, rar_files, force):
        """
        Extracts RAR files

        :param path: Path to look for files in
        :param rar_files: Names of RAR files
        :param force: process currently processing items
        :return: List of unpacked file names
        """

        unpacked_dirs = []

        if sickrage.app.config.unpack == 1 and rar_files:
            self.log("Packed Releases detected: {0}".format(rar_files), sickrage.app.log.DEBUG)
            for archive in rar_files:
                failure = None
                rar_handle = None
                try:
                    archive_path = os.path.join(path, archive)
                    if self.already_postprocessed(path, archive, force):
                        self.log("Archive file already post-processed, extraction skipped: {}".format
                                 (archive_path), sickrage.app.log.DEBUG)
                        continue

                    if not is_rar_file(archive_path):
                        continue

                    self.log(
                        "Checking if archive is valid and contains a video: {}".format(archive_path),
                        sickrage.app.log.DEBUG)
                    rar_handle = rarfile.RarFile(archive_path)
                    if rar_handle.needs_password():
                        # TODO: Add support in settings for a list of passwords to try here with rar_handle.set_password(x)
                        self.log('Archive needs a password, skipping: {0}'.format(archive_path))
                        continue

                    # If there are no video files in the rar, don't extract it
                    rar_media_files = list(filter(is_media_file, rar_handle.namelist()))
                    if not rar_media_files:
                        continue

                    rar_release_name = archive.rpartition('.')[0]

                    # Choose the directory we'll unpack to:
                    if sickrage.app.config.unpack_dir and os.path.isdir(sickrage.app.config.unpack_dir):
                        unpack_base_dir = sickrage.app.config.unpack_dir
                    else:
                        unpack_base_dir = path
                        if sickrage.app.config.unpack_dir:  # Let user know if we can't unpack there
                            self.log('Unpack directory cannot be verified. Using {}'.format(path),
                                     sickrage.app.log.DEBUG)

                    # Fix up the list for checking if already processed
                    rar_media_files = [os.path.join(unpack_base_dir, rar_release_name, rar_media_file) for
                                       rar_media_file in
                                       rar_media_files]

                    skip_rar = False
                    for rar_media_file in rar_media_files:
                        check_path, check_file = os.path.split(rar_media_file)
                        if self.already_postprocessed(check_path, check_file, force):
                            self.log(
                                "Archive file already post-processed, extraction skipped: {0}".format
                                (rar_media_file), sickrage.app.log.DEBUG)
                            skip_rar = True
                            break

                    if skip_rar:
                        continue

                    rar_extract_path = os.path.join(unpack_base_dir, rar_release_name)
                    self.log("Unpacking archive: {0}".format(archive), sickrage.app.log.DEBUG)
                    rar_handle.extractall(path=rar_extract_path)
                    unpacked_dirs.append(rar_extract_path)

                except rarfile.RarCRCError:
                    failure = ('Archive Broken', 'Unpacking failed because of a CRC error')
                except rarfile.RarWrongPassword:
                    failure = ('Incorrect RAR Password', 'Unpacking failed because of an Incorrect Rar Password')
                except rarfile.PasswordRequired:
                    failure = ('Rar is password protected', 'Unpacking failed because it needs a password')
                except rarfile.RarOpenError:
                    failure = ('Rar Open Error, check the parent folder and destination file permissions.',
                               'Unpacking failed with a File Open Error (file permissions?)')
                except rarfile.RarExecError:
                    failure = ('Invalid Rar Archive Usage',
                               'Unpacking Failed with Invalid Rar Archive Usage. Is unrar installed and on the system '
                               'PATH?')
                except rarfile.BadRarFile:
                    failure = ('Invalid Rar Archive', 'Unpacking Failed with an Invalid Rar Archive Error')
                except rarfile.NeedFirstVolume:
                    continue
                except (Exception, rarfile.Error) as e:
                    failure = (e, 'Unpacking failed')
                finally:
                    if rar_handle:
                        del rar_handle

                if failure:
                    self.log('Failed to extract the archive {}: {}'.format(archive, failure[0]),
                             sickrage.app.log.WARNING)
                    self.missed_files.append('{} : Unpacking failed: {}'.format(archive, failure[1]))
                    self.result = False
                    continue

        return unpacked_dirs

    def already_postprocessed(self, dirName, videofile, force):
        """
        Check if we already post processed a file

        :param dirName: Directory a file resides in
        :param videofile: File name
        :param force: Force checking when already checking (currently unused)
        :return:
        """
        if force:
            return False

        session = sickrage.app.main_db.session()

        # Avoid processing the same dir again if we use a process method <> move
        if session.query(MainDB.TVEpisode).filter(or_(MainDB.TVEpisode.release_name.contains(dirName), MainDB.TVEpisode.release_name.contains(videofile))).count() > 0:
            return True

        # Needed if we have downloaded the same episode @ different quality
        # But we need to make sure we check the history of the episode we're going to PP, and not others
        np = NameParser(dirName)
        try:
            parse_result = np.parse(dirName)
        except:
            parse_result = False

        for h in session.query(MainDB.History).filter(MainDB.History.resource.endswith(videofile)):
            for e in session.query(MainDB.TVEpisode).filter_by(showid=h.showid, season=h.season, episode=h.episode).filter(MainDB.TVEpisode.status.in_(Quality.DOWNLOADED)):
                if parse_result and (parse_result.indexer_id and parse_result.episode_numbers and parse_result.season_number):
                    if e.showid == int(parse_result.indexer_id) and e.season == int(parse_result.season_number and e.episode) == int(
                            parse_result.episode_numbers[0]):
                        return True
                else:
                    return True

        # Checks for processed file marker
        if os.path.isfile(os.path.join(dirName, videofile + '.sr_processed')):
            return True

        return False

    def process_media(self, processPath, videoFiles, nzbName, process_method, force, is_priority):
        """
        Postprocess mediafiles

        :param processPath: Path to postprocess in
        :param videoFiles: Filenames to look for and postprocess
        :param nzbName: Name of NZB file related
        :param process_method: auto/manual
        :param force: Postprocess currently postprocessing file
        :param is_priority: Boolean, is this a priority download
        """

        processor = None
        for cur_video_file in videoFiles:
            cur_video_file_path = os.path.join(processPath, cur_video_file)

            if self.already_postprocessed(processPath, cur_video_file, force):
                self.log("Skipping already processed file: {0}".format(cur_video_file), sickrage.app.log.DEBUG)
                continue

            try:
                processor = post_processor.PostProcessor(cur_video_file_path, nzbName, process_method, is_priority)
                self.result = processor.process()
                process_fail_message = ""
            except EpisodePostProcessingFailedException as e:
                self.result = False
                process_fail_message = "{}".format(e)

            if processor:
                self._output.append(processor.log)

            if self.result:
                self.log("Processing succeeded for " + cur_video_file_path)
            else:
                self.log("Processing failed for {0}: {1}".format(cur_video_file_path, process_fail_message),
                         sickrage.app.log.WARNING)
                self.missed_files.append(
                    "{0} : Processing failed: {1}".format(cur_video_file_path, process_fail_message))
                self.succeeded = False

    def process_failed(self, dirName, nzbName):
        """Process a download that did not complete correctly"""

        try:
            processor = failed_processor.FailedProcessor(dirName, nzbName)
            self.result = processor.process()
            process_fail_message = ""
        except FailedPostProcessingFailedException as e:
            processor = None
            self.result = False
            process_fail_message = e

        if processor:
            self._output.append(processor.log)

        if sickrage.app.config.delete_failed and self.result:
            if self.delete_folder(dirName, check_empty=False):
                self.log("Deleted folder: " + dirName, sickrage.app.log.DEBUG)

        if self.result:
            self.log(
                "Failed Download Processing succeeded: (" + str(nzbName) + ", " + dirName + ")")
        else:
            self.log("Failed Download Processing failed: ({}, {}): {}"
                     .format(nzbName, dirName, process_fail_message),
                     sickrage.app.log.WARNING)



from __future__ import absolute_import, division, print_function

import requests
from trakt.interfaces.base import Interface
from trakt.mapper.summary import SummaryMapper


class MoviesInterface(Interface):
    path = 'movies'

    def get(self, id, extended=None, **kwargs):
        response = self.http.get(str(id), query={
            'extended': extended
        })

        items = self.get_data(response, **kwargs)

        if isinstance(items, requests.Response):
            return items

        # Parse response
        return SummaryMapper.movie(self.client, items)

    def trending(self, extended=None, **kwargs):
        response = self.http.get('trending', query={
            'extended': extended
        })

        items = self.get_data(response, **kwargs)

        if isinstance(items, requests.Response):
            return items

        return SummaryMapper.movies(self.client, items)



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
# Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.


import datetime
import threading

import sickrage
from sickrage.core.common import Quality, SNATCHED, SNATCHED_BEST, SNATCHED_PROPER
from sickrage.core.databases.main import MainDB
from sickrage.core.queues.search import FailedSearchTask
from sickrage.core.tv.show.helpers import find_show
from sickrage.core.tv.show.history import FailedHistory


class FailedSnatchSearcher(object):
    def __init__(self):
        self.name = "FAILEDSNATCHSEARCHER"
        self.lock = threading.Lock()
        self.running = False

    def task(self, force=False):
        """
        Runs the failed searcher, queuing selected episodes for search that have failed to snatch
        :param force: Force search
        """
        if self.running or not sickrage.app.config.use_failed_snatcher and not force:
            return

        try:
            self.running = True

            # set thread name
            threading.currentThread().setName(self.name)

            # trim failed download history
            FailedHistory.trim_history()

            sickrage.app.log.info("Searching for failed snatches")

            failed_snatches = False

            for snatched_episode_obj in [x for x in self.snatched_episodes() if (x.showid, x.season, x.episode) not in self.downloaded_releases()]:
                show_object = find_show(snatched_episode_obj.showid)
                episode_object = show_object.get_episode(snatched_episode_obj.season, snatched_episode_obj.episode)
                if episode_object.show.paused:
                    continue

                cur_status, cur_quality = Quality.split_composite_status(episode_object.status)
                if cur_status not in {SNATCHED, SNATCHED_BEST, SNATCHED_PROPER}:
                    continue

                sickrage.app.search_queue.put(FailedSearchTask(episode_object.showid, episode_object.season, episode_object.episode, True))

                failed_snatches = True

            if not failed_snatches:
                sickrage.app.log.info("No failed snatches found")
        finally:
            self.running = False

    def snatched_episodes(self):
        session = sickrage.app.main_db.session()
        return (x for x in
                session.query(MainDB.History).filter(MainDB.History.action.in_(Quality.SNATCHED + Quality.SNATCHED_BEST + Quality.SNATCHED_PROPER)) if
                24 >= int((datetime.datetime.now() - datetime.datetime.fromordinal(x.date)).total_seconds() / 3600) >= sickrage.app.config.failed_snatch_age)

    def downloaded_releases(self):
        session = sickrage.app.main_db.session()
        return ((x.showid, x.season, x.episode) for x in session.query(MainDB.History).filter(MainDB.History.action.in_(Quality.DOWNLOADED)))



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
import datetime

from sqlalchemy import Column, Integer, Text, ForeignKeyConstraint, String, DateTime, Boolean, Index, Date, BigInteger, func, literal_column
from sqlalchemy.ext.declarative import as_declarative
from sqlalchemy.orm import relationship

from sickrage.core import common
from sickrage.core.common import SearchFormats
from sickrage.core.databases import SRDatabase, SRDatabaseBase


@as_declarative()
class MainDBBase(SRDatabaseBase):
    pass


class MainDB(SRDatabase):
    def __init__(self, db_type, db_prefix, db_host, db_port, db_username, db_password):
        super(MainDB, self).__init__('main', 12, db_type, db_prefix, db_host, db_port, db_username, db_password)
        MainDBBase.metadata.create_all(self.engine)
        for model in MainDBBase._decl_class_registry.values():
            if hasattr(model, '__tablename__'):
                self.tables[model.__tablename__] = model

    def cleanup(self):
        def remove_duplicate_shows():
            session = self.session()

            duplicates = session.query(self.TVShow, func.count(self.TVShow.indexer_id).label('count')). \
                group_by(self.TVShow.indexer_id). \
                having(literal_column('count') > 1)

            for cur_duplicate in duplicates:
                session.query(self.TVShow).filter_by(showid=cur_duplicate['indexer_id']).\
                    limit(int(cur_duplicate['count']) - 1).delete()
                session.commit()

        def remove_duplicate_episodes():
            session = self.session()

            duplicates = session.query(self.TVEpisode, func.count(self.TVEpisode.showid).label('count')). \
                group_by(self.TVEpisode.showid, self.TVEpisode.season, self.TVEpisode.episode). \
                having(literal_column('count') > 1)

            for cur_duplicate in duplicates:
                session.query(self.TVEpisode). \
                    filter_by(showid=cur_duplicate['showid'], season=cur_duplicate['season'], episode=cur_duplicate['episode']). \
                    limit(int(cur_duplicate['count']) - 1).delete()
                session.commit()

        remove_duplicate_shows()
        remove_duplicate_episodes()

    class TVShow(MainDBBase):
        __tablename__ = 'tv_shows'

        indexer_id = Column(Integer, index=True, primary_key=True)
        indexer = Column(Integer, index=True, primary_key=True)
        name = Column(Text, default='')
        location = Column(Text, default='')
        network = Column(Text, default='')
        genre = Column(Text, default='')
        overview = Column(Text, default='')
        classification = Column(Text, default='Scripted')
        runtime = Column(Integer, default=0)
        quality = Column(Integer, default=-1)
        airs = Column(Text, default='')
        status = Column(Text, default='')
        flatten_folders = Column(Boolean, default=0)
        paused = Column(Boolean, default=0)
        search_format = Column(Integer, default=SearchFormats.STANDARD)
        anime = Column(Boolean, default=0)
        subtitles = Column(Boolean, default=0)
        dvdorder = Column(Boolean, default=0)
        skip_downloaded = Column(Boolean, default=0)
        startyear = Column(Integer, default=0)
        lang = Column(Text, default='')
        imdb_id = Column(Text, default='')
        rls_ignore_words = Column(Text, default='')
        rls_require_words = Column(Text, default='')
        default_ep_status = Column(Integer, default=common.SKIPPED)
        sub_use_sr_metadata = Column(Boolean, default=0)
        notify_list = Column(Text, default='')
        search_delay = Column(Integer, default=0)
        scene_exceptions = Column(Text, default='')
        last_scene_exceptions_refresh = Column(Integer, default=0)
        last_update = Column(Integer, default=datetime.datetime.now().toordinal())
        last_refresh = Column(Integer, default=datetime.datetime.now().toordinal())
        last_backlog_search = Column(Integer, default=0)
        last_proper_search = Column(Integer, default=0)

        episodes = relationship('TVEpisode', uselist=True, backref='tv_shows', lazy='dynamic')
        imdb_info = relationship('IMDbInfo', uselist=False, backref='tv_shows')

    class TVEpisode(MainDBBase):
        __tablename__ = 'tv_episodes'
        __table_args__ = (
            ForeignKeyConstraint(['showid', 'indexer'], ['tv_shows.indexer_id', 'tv_shows.indexer']),
            Index('idx_showid_indexer', 'showid', 'indexer'),
            Index('idx_showid_indexerid', 'showid', 'indexer_id'),
            Index('idx_sta_epi_air', 'status', 'episode', 'airdate'),
            Index('idx_sea_epi_sta_air', 'season', 'episode', 'status', 'airdate'),
            Index('idx_indexer_id_airdate', 'indexer_id', 'airdate'),
        )

        showid = Column(Integer, index=True, primary_key=True)
        indexer_id = Column(Integer, default=0)
        indexer = Column(Integer, index=True, primary_key=True)
        season = Column(Integer, index=True, primary_key=True)
        episode = Column(Integer, index=True, primary_key=True)
        scene_season = Column(Integer, default=0)
        scene_episode = Column(Integer, default=0)
        name = Column(Text, default='')
        description = Column(Text, default='')
        subtitles = Column(Text, default='')
        subtitles_searchcount = Column(Integer, default=0)
        subtitles_lastsearch = Column(Integer, default=0)
        airdate = Column(Date, default=datetime.datetime.min)
        hasnfo = Column(Boolean, default=False)
        hastbn = Column(Boolean, default=False)
        status = Column(Integer, default=common.UNKNOWN)
        location = Column(Text, default='')
        file_size = Column(BigInteger, default=0)
        release_name = Column(Text, default='')
        is_proper = Column(Boolean, default=False)
        absolute_number = Column(Integer, default=0)
        scene_absolute_number = Column(Integer, default=0)
        version = Column(Integer, default=-1)
        release_group = Column(Text, default='')

        show = relationship('TVShow', uselist=False, backref='tv_episodes')

    class IMDbInfo(MainDBBase):
        __tablename__ = 'imdb_info'
        __table_args__ = (
            ForeignKeyConstraint(['indexer_id'], ['tv_shows.indexer_id']),
        )

        indexer_id = Column(Integer, primary_key=True)
        imdb_id = Column(String(10), index=True, unique=True)
        rated = Column(Text)
        title = Column(Text)
        production = Column(Text)
        website = Column(Text)
        writer = Column(Text)
        actors = Column(Text)
        type = Column(Text)
        votes = Column(Text, nullable=False)
        seasons = Column(Text)
        poster = Column(Text)
        director = Column(Text)
        released = Column(Text)
        awards = Column(Text)
        genre = Column(Text, nullable=False)
        rating = Column(Text, nullable=False)
        language = Column(Text)
        country = Column(Text)
        runtime = Column(Text)
        metascore = Column(Text)
        year = Column(Text)
        plot = Column(Text)
        last_update = Column(Integer, nullable=False)

    class XEMRefresh(MainDBBase):
        __tablename__ = 'xem_refresh'

        indexer_id = Column(Integer, primary_key=True)
        indexer = Column(Integer, primary_key=True)
        last_refreshed = Column(Integer, nullable=False)

    class SceneNumbering(MainDBBase):
        __tablename__ = 'scene_numbering'

        indexer = Column(Integer, primary_key=True)
        indexer_id = Column(Integer, primary_key=True)
        season = Column(Integer, primary_key=True)
        episode = Column(Integer, primary_key=True)
        scene_season = Column(Integer, nullable=False)
        scene_episode = Column(Integer, nullable=False)
        absolute_number = Column(Integer, nullable=False)
        scene_absolute_number = Column(Integer, nullable=False)

    class IndexerMapping(MainDBBase):
        __tablename__ = 'indexer_mapping'

        indexer_id = Column(Integer, primary_key=True)
        indexer = Column(Integer, primary_key=True)
        mindexer_id = Column(Integer, nullable=False)
        mindexer = Column(Integer, primary_key=True)

    class Blacklist(MainDBBase):
        __tablename__ = 'blacklist'

        id = Column(Integer, primary_key=True)
        show_id = Column(Integer, nullable=False)
        keyword = Column(Text, nullable=False)

    class Whitelist(MainDBBase):
        __tablename__ = 'whitelist'

        id = Column(Integer, primary_key=True)
        show_id = Column(Integer, nullable=False)
        keyword = Column(Text, nullable=False)

    class History(MainDBBase):
        __tablename__ = 'history'

        id = Column(Integer, primary_key=True)
        showid = Column(Integer, nullable=False)
        season = Column(Integer, nullable=False)
        episode = Column(Integer, nullable=False)
        resource = Column(Text, nullable=False)
        action = Column(Integer, nullable=False)
        version = Column(Integer, default=-1)
        provider = Column(Text, nullable=False)
        date = Column(DateTime, nullable=False)
        quality = Column(Integer, nullable=False)
        release_group = Column(Text, nullable=False)

    class FailedSnatchHistory(MainDBBase):
        __tablename__ = 'failed_snatch_history'

        id = Column(Integer, primary_key=True)
        date = Column(DateTime, nullable=False)
        size = Column(Integer, nullable=False)
        release = Column(Text, nullable=False)
        provider = Column(Text, nullable=False)
        showid = Column(Integer, nullable=False)
        season = Column(Integer, nullable=False)
        episode = Column(Integer, nullable=False)
        old_status = Column(Integer, nullable=False)

    class FailedSnatch(MainDBBase):
        __tablename__ = 'failed_snatches'

        id = Column(Integer, primary_key=True)
        release = Column(Text, nullable=False)
        size = Column(Integer, nullable=False)
        provider = Column(Text, nullable=False)



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.


import re
import time
from base64 import b16encode, b32decode
from hashlib import sha1

from bencode3 import bdecode, BencodeError, bencode

import sickrage
from sickrage.core.websession import WebSession

_clients = {
    'utorrent': 'uTorrentAPI',
    'transmission': 'TransmissionAPI',
    'deluge': 'DelugeAPI',
    'deluged': 'DelugeDAPI',
    'download_station': 'DownloadStationAPI',
    'rtorrent': 'rTorrentAPI',
    'qbittorrent': 'QBittorrentAPI',
    'mlnet': 'mlnetAPI',
    'putio': 'PutioAPI',
}


class GenericClient(object):
    def __init__(self, name, host=None, username=None, password=None):
        self.name = name
        self.username = sickrage.app.config.torrent_username if not username else username
        self.password = sickrage.app.config.torrent_password if not password else password
        self.host = sickrage.app.config.torrent_host if not host else host
        self.rpcurl = sickrage.app.config.torrent_rpcurl

        self.url = None
        self.auth = None
        self.last_time = time.time()

        self.session = WebSession(cache=False)

        self._response = None

    @property
    def response(self):
        return self._response

    @response.setter
    def response(self, value):
        self._response = value

    def _request(self, method='get', params=None, data=None, *args, **kwargs):
        if time.time() > self.last_time + 1800 or not self.auth:
            self.last_time = time.time()
            self._get_auth()

        sickrage.app.log.debug(
            '{name}: Requested a {method} connection to {url} with'
            ' params: {params} Data: {data}'.format(
                name=self.name,
                method=method.upper(),
                url=self.url,
                params=params,
                data=str(data)[0:99] + '...' if len(str(data)) > 102 else str(data)
            )
        )

        if not self.auth:
            sickrage.app.log.warning(self.name + ': Authentication Failed')
            return False

        self.response = self.session.request(method.upper(), self.url,
                                             params=params, data=data, auth=(self.username, self.password), timeout=120, verify=False, *args, **kwargs)

        if not self.response or not self.response.text:
            return False

        sickrage.app.log.debug('{name}: Response to {method} request is {response}'.format(
            name=self.name,
            method=method.upper(),
            response=self.response.text
        ))

        return True

    def _get_auth(self):
        """
        This should be overridden and should return the auth_id needed for the client
        """
        return None

    def test_authentication(self):
        # verify valid url
        self.response = self.session.get(self.url or self.host, timeout=120, verify=False)
        if self.response is None:
            return False, 'Error: Unable to connect to ' + self.name

        # verify auth
        if self._get_auth():
            return True, 'Success: Connected and Authenticated'
        return False, 'Error: Unable to get ' + self.name + ' Authentication, check your config!'


class TorrentClient(GenericClient):
    def _add_torrent_uri(self, result):
        """
        This should be overridden should return the True/False from the client
        when a torrent is added via url (magnet or .torrent link)
        """
        return False

    def _add_torrent_file(self, result):
        """
        This should be overridden should return the True/False from the client
        when a torrent is added via result.content (only .torrent file)
        """
        return False

    def _set_torrent_label(self, result):
        """
        This should be overridden should return the True/False from the client
        when a torrent is set with label
        """
        return True

    def _set_torrent_ratio(self, result):
        """
        This should be overridden should return the True/False from the client
        when a torrent is set with ratio
        """
        return True

    def _set_torrent_seed_time(self, result):
        """
        This should be overridden should return the True/False from the client
        when a torrent is set with a seed time
        """
        return True

    def _set_torrent_priority(self, result):
        """
        This should be overriden should return the True/False from the client
        when a torrent is set with result.priority (-1 = low, 0 = normal, 1 = high)
        """
        return True

    def _set_torrent_path(self, torrent_path):
        """
        This should be overridden should return the True/False from the client
        when a torrent is set with path
        """
        return True

    def _set_torrent_pause(self, result):
        """
        This should be overridden should return the True/False from the client
        when a torrent is set with pause
        """
        return True

    @staticmethod
    def _get_torrent_hash(result):
        if result.url.startswith('magnet'):
            result.hash = re.findall(r'urn:btih:([\w]{32,40})', result.url)[0]
            if len(result.hash) == 32:
                result.hash = b16encode(b32decode(result.hash)).lower()
        else:
            if not result.content:
                sickrage.app.log.warning('Torrent without content')
                raise Exception('Torrent without content')

            try:
                torrent_bdecode = bdecode(result.content)
            except BencodeError:
                sickrage.app.log.warning('Unable to bdecode torrent')
                sickrage.app.log.debug('Torrent bencoded data: %r' % result.content)
                raise

            try:
                info = torrent_bdecode["info"]
            except Exception:
                sickrage.app.log.warning('Unable to find info field in torrent')
                raise

            result.hash = sha1(bencode(info)).hexdigest()

        return result

    def send_torrent(self, result):

        r_code = False

        sickrage.app.log.debug('Calling ' + self.name + ' Client')

        try:
            if not self._get_auth():
                sickrage.app.log.warning(self.name + ': Authentication Failed')
                return r_code

            # Sets per provider seed ratio
            result.ratio = result.provider.seed_ratio

            # lazy fix for now, I'm sure we already do this somewhere else too
            result = self._get_torrent_hash(result)

            # convert to magnetic url if result has info hash and is not a private provider
            if sickrage.app.config.torrent_file_to_magnet:
                if result.hash and not result.provider.private and not result.url.startswith('magnet'):
                    result.url = "magnet:?xt=urn:btih:{}".format(result.hash)

            if result.url.startswith('magnet'):
                r_code = self._add_torrent_uri(result)
            else:
                r_code = self._add_torrent_file(result)

            if not r_code:
                sickrage.app.log.warning(self.name + ': Unable to send Torrent')
                return False

            if not self._set_torrent_pause(result):
                sickrage.app.log.warning(self.name + ': Unable to set the pause for Torrent')

            if not self._set_torrent_label(result):
                sickrage.app.log.warning(self.name + ': Unable to set the label for Torrent')

            if not self._set_torrent_ratio(result):
                sickrage.app.log.warning(self.name + ': Unable to set the ratio for Torrent')

            if not self._set_torrent_seed_time(result):
                sickrage.app.log.warning(self.name + ': Unable to set the seed time for Torrent')

            if not self._set_torrent_path(result):
                sickrage.app.log.warning(self.name + ': Unable to set the path for Torrent')

            if result.priority != 0 and not self._set_torrent_priority(result):
                sickrage.app.log.warning(self.name + ': Unable to set priority for Torrent')
        except Exception as e:
            sickrage.app.log.warning(self.name + ': Failed Sending Torrent')
            sickrage.app.log.debug(self.name + ': Exception raised when sending torrent: {}. Error: {}'.format(result, e))
            return r_code

        return r_code


class NZBClient(GenericClient):
    def _add_nzb_uri(self, result):
        """
        This should be overridden should return the True/False from the client
        when a torrent is added via url (magnet or .torrent link)
        """
        return False

    def _add_nzb_file(self, result):
        """
        This should be overridden should return the True/False from the client
        when a torrent is added via result.content (only .torrent file)
        """
        return False

    def send_nzb(self, result):
        if result.resultType == 'nzb':
            return self._add_nzb_uri(result)
        elif result.resultType == 'nzbdata':
            return self._add_nzb_file(result)


def get_client_module(name, client_type):
    return __import__("{}.{}.{}".format(__name__, client_type, name.lower()), fromlist=list(_clients.keys()))


def get_client_instance(name, client_type):
    return getattr(get_client_module(name, client_type), _clients[name])



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
# Git: https://git.sickrage.ca/SiCKRAGE/sickrage
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.


import re

import sickrage
from sickrage.core.caches.tv_cache import TVCache
from sickrage.core.helpers import bs4_parser, convert_size
from sickrage.providers import TorrentProvider


class TORRENTZProvider(TorrentProvider):
    def __init__(self):

        super(TORRENTZProvider, self).__init__("Torrentz", 'https://torrentz2.eu', False)
        self.confirmed = True

        self.minseed = None
        self.minleech = None

        self._urls.update({
            'verified': '{base_url}/feed_verified'.format(**self._urls),
            'feed': '{base_url}/feed'.format(**self._urls)
        })

        self.cache = TVCache(self, min_time=15)

    @staticmethod
    def _split_description(description):
        match = re.findall(r'[0-9]+', description)
        return int(match[0]) * 1024 ** 2, int(match[1]), int(match[2])

    def search(self, search_strings, age=0, show_id=None, season=None, episode=None, **kwargs):
        results = []

        for mode in search_strings:
            sickrage.app.log.debug('Search Mode: {}'.format(mode))
            for search_string in search_strings[mode]:
                search_params = {'f': search_string}
                search_url = self.urls['feed']

                if mode != 'RSS':
                    sickrage.app.log.debug('Search string: {}'.format(search_string))

                resp = self.session.get(self.urls['search'], params=search_params)
                if not resp or not resp.text:
                    sickrage.app.log.debug("No data returned from provider")
                    continue

                results += self.parse(resp.text, mode)

        return results

    def parse(self, data, mode, **kwargs):
        """
        Parse search results from data
        :param data: response data
        :param mode: search mode
        :return: search results
        """

        results = []

        if not data.startswith('<?xml'):
            sickrage.app.log.info('Expected xml but got something else, is your mirror failing?')
            return results

        with bs4_parser(data) as parser:
            for item in parser('item'):
                try:
                    if item.category and 'tv' not in item.category.get_text(strip=True).lower():
                        continue

                    title = item.title.get_text(strip=True)
                    t_hash = item.guid.get_text(strip=True).rsplit('/', 1)[-1]
                    if not all([title, t_hash]):
                        continue

                    download_url = "magnet:?xt=urn:btih:" + t_hash + "&dn=" + title
                    torrent_size, seeders, leechers = self._split_description(item.find('description').text)
                    size = convert_size(torrent_size, -1)

                    results += [
                        {'title': title, 'link': download_url, 'size': size, 'seeders': seeders, 'leechers': leechers}
                    ]

                    if mode != 'RSS':
                        sickrage.app.log.debug("Found result: {}".format(title))
                except Exception:
                    sickrage.app.log.error("Failed parsing provider.")

        return results
